cdef class SoundBuffer:
    cdef public int samplerate
    cdef public int channels
    cdef public double[:,:] frames

    cdef SoundBuffer _adsr(self, double attack, double decay, double sustain, double release)
    cdef void _dub(self, SoundBuffer sound, int framepos=*)
    cdef void _fill(self, double[:,:] frames)
    cdef double[:,:] _speed(self, double speed)

cdef class RingBuffer:
    cdef public int length
    cdef public int samplerate
    cdef public int channels
    cdef public int write_head
    cdef public double[:,:] frames
    cdef double[:,:] copyout

    cdef double[:,:] _read(self, int length, int offset)
    cdef void _write(self, double[:,:] frames)



""" Some helpers for building and transforming onset lists
"""

from . import wavetables
from . import interpolation
from . import dsp
import pyparsing

REST_SYMBOLS = set(('0', '.', ' ', '-', 0, False))

CLAVE = {
    'son':      'x..x..x...x.x...', 
    'rumba':    'x..x...x..x.x...', 
    'bell':     'x..x..xx..x.x..x', 
    'tresillo': 'x..x..x.', 
}


def pattern(
    p,              # Pattern
    beat=0.25,      # Beat length in seconds
    length=1,       # Output length in seconds 
    bpm=None,       # Tempo in beats per minute
    hz=None,        # Or tempo in hz (overrides bpm)
    swing=0,        # MPC swing amount 0-1
    lfo=None,       # Apply lfo tempo modulation across pattern (string, iterable, soundbuffer, etc)
    lfo_tempo=None, # Lfo frequency in BPM string or Hz
    lfo_depth=None, # Fractional multiplier from original tempo
    delay=False     # Fixed delay in seconds added to each onset
):
    """ Onsets from ascii
    """

    if bpm is not None:
        bpm = bpm if bpm > 0 else np.nextafter(0, 1)
        beat = 60 / bpm
    elif hz is not None:
        hz = hz if hz > 0 else np.nextafter(0, 1)
        beat = 1 / hz  

    positions = topositions(p, beat, length)
    positions = onsetswing(onsets, swing, beat)

    if delay:
        positions = [ pos + delay for pos in positions ]

    return positions

def topositions(p, beat=0.25, length=1):
    pos = 0
    count = 0
    div = 1
    out = []

    while pos < length:
        index = count % len(p)
        event = p[index]

        if event in REST_SYMBOLS:
            count += 1
            pos += beat
            continue

        elif event == '[':
            end = p.find(']', index) - 1
            div = len(p[index : end])
            beat /= div
            count += 1
            continue

        elif event == ']':
            beat *= div
            count += 1
            continue

        out += [ pos ]
        pos += beat
        count += 1

    return out

def render(pattern, callback, beat=0.25, length=1, truncate=False):
    out = dsp.buffer(length=length)

    positions = topositions(pattern, beat, length)

    for pos in positions:
        clang = callback(pos/length)
        if truncate and pos + clang.dur > length:
            clang = clang.cut(0, length - pos)
        out.dub(clang, pos)

    return out


def normalize_pattern(pattern, reverse=False, rotate=None):
    pattern = [ 0 if tick in REST_SYMBOLS or not tick else 1 for tick in pattern ]
    if reverse:
        pattern = [ p for p in reversed(pattern) ]

    if rotate:
        pattern = [ 0 for _ in range(rotate) ] + pattern[:-rotate]

    return pattern

def onsetswing(onsets, amount, beat):
    """ Add MPC-style swing to a list of onsets.
        Amount is a value between 0 and 1, which 
        maps to a swing amount between 0% and 75%.
        Every odd onset will be delayed by
            
            beat length * swing amount

        This will only really work like MPC swing 
        when there is a run of notes, as rests 
        are not represented in onset lists (and 
        MPC swing just delays the upbeats).
    """
    if amount <= 0:
        return onsets

    delay_amount = int(beat * amount * 0.75)
    for i, onset in enumerate(onsets):
        if i % 2 == 1:
            onsets[i] += delay_amount

    return onsets


def pgen(numbeats, div=1, offset=0, reps=None, reverse=False):
    """ Pattern creation helper
    """
    pat = [ 1 if tick == 0 else 0 for tick in range(div) ]
    pat = [ pat[i % len(pat)] for i in range(numbeats) ]

    if offset > 0:
        pat = [ 0 for _ in range(offset) ] + pat[:-offset]

    if reps is not None:
        pat *= reps

    if reverse:
        pat = [ p for p in reversed(pat) ]

    return pat

def onsets(pattern, beat=0.2, length=None, start=0):
    length = length or len(pattern)
    grid = [ beat * i + start for i in range(length) ]
    pattern = [ pattern[i % len(pattern)] for i in range(length) ]

    out = []

    for i, tick in enumerate(pattern):
        if tick not in REST_SYMBOLS:
            pos = grid[i % len(grid)]
            out += [ pos ]
        
    return out

def eu(length, numbeats, offset=0, reps=None, reverse=False):
    """ A euclidian pattern generator

        Length 6, numbeats 3
        >>> rhythm.eu(6, 3)
        [1, 0, 1, 0, 1, 0]

        Length 6, numbeats 3, offset 1
        >>> rhythm.eu(6, 3, 1)
        [0, 1, 0, 1, 0, 1]
    """
    pulses = [ 1 for pulse in range(numbeats) ]
    pauses = [ 0 for pause in range(length - numbeats) ]

    position = 0
    while len(pauses) > 0:
        try:
            index = pulses.index(1, position)
            pulses.insert(index + 1, pauses.pop(0))
            position = index + 1
        except ValueError:
            position = 0

    pattern = rotate(pulses, offset+len(pulses))

    if reps:
        pattern *= reps

    if reverse:
        pattern = [ p for p in reversed(pattern) ]

    return pattern

def curve(numbeats=16, wintype=None, length=44100, reverse=False, wavetable=None):
    """ FIXME -- use seconds, rework api
    """
    wintype = wintype or dsp.RND

    if wavetable is None:
        win = wavetables.window(wintype, numbeats * 2)
    else:
        win = interpolation.linear(wavetable, numbeats)

    if reverse:
        win = win[numbeats:]
    else:
        win = win[:numbeats]

    return [ int(onset * length) for onset in win ]

def rotate(pattern, offset=0):
    """ Rotate a pattern list by a given offset
    """
    return pattern[-offset % len(pattern):] + pattern[:-offset % len(pattern)]

def repeat(onsets, reps):
    """ Repeat a sequence of onsets a given number of times
    """
    out = []
    total = sum(onsets)
    for rep in range(reps):
        out += [ onset + (rep * total) for onset in onsets ]

    return out




# cython: language_level=3

import collections
import random
import numbers

import soundfile
import numpy as np
import re

from libc.stdlib cimport malloc, realloc, calloc, free
from libc.stdlib cimport rand
from libc cimport math

from . cimport interpolation
from .soundbuffer cimport SoundBuffer

cdef int SINE = 0
cdef int COS = 1
cdef int TRI = 2
cdef int SAW = 3
cdef int PHASOR = SAW
cdef int RSAW = 4
cdef int HANN = 5
cdef int HAMM = 6
cdef int BLACK = 7
cdef int BLACKMAN = 7
cdef int BART = 8
cdef int BARTLETT = 8
cdef int KAISER = 9
cdef int SQUARE = 10
cdef int RND = 11

cdef dict wtype_flags = {
    'sine': SINE, 
    'cos': COS, 
    'tri': TRI, 
    'saw': SAW, 
    'phasor': PHASOR, 
    'rsaw': RSAW, 
    'hann': HANN, 
    'hamm': HAMM, 
    'black': BLACK, 
    'blackman': BLACKMAN, 
    'bart': BART, 
    'bartlett': BARTLETT, 
    'kaiser': KAISER, 
    'rnd': RND, 
}

cdef int LEN_WINDOWS = 9
cdef int *ALL_WINDOWS = [
            SINE, 
            TRI, 
            SAW,
            RSAW,
            HANN,
            HAMM,
            BLACK,
            BART,
            KAISER
        ]

cdef int LEN_WAVETABLES = 6
cdef int *ALL_WAVETABLES = [
            SINE, 
            COS,
            TRI,
            SAW,
            RSAW,
            SQUARE
        ]

cdef double SQUARE_DUTY = 0.5

SEGMENT_RE = re.compile('(?P<length>0?\.?\d+)?,?(?P<wtype>\w+),?(?P<start>0?\.?\d+)?-?(?P<end>0?\.?\d+)?')

cdef int wtypentf(str wtype_name):
    return wtype_flags[wtype_name]


cdef class Wavetable:
    def __cinit__(self, list values=None, int initwt=-1, int initwin=-1, int length=-1):
        if length < 0:
            length = 4096

        if values is not None:
            self.data = np.array(values)
        elif initwt >= 0:
            self.data = _wavetable(initwt, length)
        elif initwin >= 0:
            self.data = _window(initwin, length)
        else:
            self.data = None


    #############################################
    # (+) Addition & concatenation operator (+) #
    #############################################
    def __add__(self, value):
        """ Add a number to every sample in this SoundBuffer or concatenate with another 
            SoundBuffer or iterable with compatible dimensions
        """
        cdef int length = len(self.data)
        cdef double[:] out = np.zeros(length)

        if isinstance(value, numbers.Real):
            out = np.add(self.data, value)
        elif isinstance(value, Wavetable):
            out = np.add(self.data, value.data)
            pass
        elif isinstance(value, SoundBuffer):
            out = np.add(self.data, value.tomono().frames)
        else:
            try:
                self.data = np.hstack((self.data, value))
            except TypeError as e:
                return NotImplemented

        return Wavetable(out)

    def __iadd__(self, value):
        """ In place add either adding number to every value without copy, or 
            directly extending internal frame buffer.
        """
        if isinstance(value, numbers.Real):
            self.data = np.add(self.data, value)
        else:
            try:
                self.data = np.hstack((self.data, value))
            except TypeError as e:
                return NotImplemented

        return self

    def __radd__(self, value):
        return self + value


    ########################
    # (&) Mix operator (&) #
    ########################
    def __and__(self, value):
        """ Mix two SoundBuffers or two 2d arrays with compatible dimensions

            >>> wt = Wavetable()
            >>> wt2 = Wavetable(list(range(20)))
            >>> wto = wt2 & wt
            >>> len(wto)
            20            
        """

        cdef double[:] out

        try:
            out = np.add(self.data, value[:len(self.data)])
            return Wavetable(out)
        except TypeError as e:
            return NotImplemented

    def __iand__(self, value):
        """ Mix in place two SoundBuffers or two 2d arrays with compatible dimensions
        """
        self.data = np.add(self.data, value[:len(self.data)])
        return self

    def __rand__(self, value):
        return self & value


    ##############
    # Truthiness #
    ##############
    def __bool__(self):
        return bool(len(self))

    def __getitem__(self, position):
        if isinstance(position, int):
            return self.data[position]
        return Wavetable(self.data[position])

    def __len__(self):
        return 0 if self.data is None else len(self.data)


    ###################################
    # (*) Multiplication operator (*) #
    ###################################
    def __mul__(self, value):
        cdef int length = len(self.data)
        cdef double[:] out = np.zeros(length)

        if isinstance(value, numbers.Real):
            out = np.multiply(self.data, value)

        elif isinstance(value, Wavetable):
            out = np.multiply(self.data, value.data)

        elif isinstance(value, SoundBuffer):
            out = np.multiply(self.data, value.frames)

        elif isinstance(value, list):
            out = np.multiply(self.data, np.array(value))

        else:
            try:
                out = np.multiply(self.data, np.array(value))
            except TypeError:
                return NotImplemented

        return Wavetable(out)

    def __imul__(self, value):
        """ Multiply a SoundBuffer in place by a number, iterable or SoundBuffer
        """
        if isinstance(value, numbers.Real):
            self.data = np.multiply(self.data, value)

        elif isinstance(value, Wavetable):
            self.data = np.multiply(self.data, value.data)

        elif isinstance(value, SoundBuffer):
            self.data = np.multiply(self.data, value.frames)

        elif isinstance(value, list):
            self.data = np.multiply(self.data, np.array(value))

        else:
            try:
                self.data = np.multiply(self.data, np.array(value))
            except TypeError:
                return NotImplemented

        return self

    def __rmul__(self, value):
        return self * value


    ################################
    # (-) Subtraction operator (-) #
    ################################
    def __sub__(self, value):
        """ Subtract a number from every sample in this SoundBuffer or  
            from another SoundBuffer or iterable with compatible dimensions
        """
        cdef double[:,:] out

        if isinstance(value, numbers.Real):
            out = np.subtract(self.data, value)
        if isinstance(value, SoundBuffer):
            out = np.subtract(self.data, value.frames)
        if isinstance(value, Wavetable):
            out = np.subtract(self.data, value.data)
        else:
            try:
                out = np.subtract(self.frames, value[:,None])
            except TypeError as e:
                return NotImplemented

        return SoundBuffer(out, channels=self.channels, samplerate=self.samplerate)

    def __isub__(self, value):
        """ In place subtract a number from every sample in this SoundBuffer or  
            from another SoundBuffer or iterable with compatible dimensions
        """
        if isinstance(value, numbers.Real):
            self.data = np.subtract(self.data, value)

        if isinstance(value, SoundBuffer):
            self.data = np.subtract(self.data, value.frames)

        if isinstance(value, Wavetable):
            self.data = np.subtract(self.data, value.data)
        else:
            try:
                self.data = np.subtract(self.data, value)
            except TypeError as e:
                return NotImplemented

    def __rsub__(self, value):
        return self - value


    def __repr__(self):
        return 'Wavetable({})'.format(self.data)

    def clip(self, minval=-1, maxval=1):
        return Wavetable(np.clip(self.data, minval, maxval))
        
    def env(self, int window_type=-1):
        """ Apply an amplitude envelope 
            to the sound of the given type.

            To modulate a sound with an arbitrary 
            iterable, simply do:

            >>> snd * iterable

            Where iterable is a list, array, or SoundBuffer with 
            the same # of channels and of any length
        """
        cdef int length = len(self)
        cdef double[:] wavetable = None
        wavetable = _window(window_type, length)

        if wavetable is None:
            raise TypeError('Invalid envelope type')

        return self * wavetable

    def max(self):
        return np.amax(self.data)
 
    def repeat(self, int reps=2):
        if reps <= 1:
            return Wavetable(self.data)

        return Wavetable(list(self.data) * reps)

    def reverse(self):
        self.data = np.flip(self.data, 0)

    def reversed(self):
        return Wavetable(np.flip(self.data, 0))

    def taper(self, double length):
        cdef int framelength = <int>(self.samplerate * length)
        return self * _adsr(len(self), framelength, 0, 1, framelength)


cdef tuple _parse_polyseg(str score, int length, int wtlength):
    """ score = '1,tri .5,sine,0-.5 sine!.25 tri,.1-.2!.5'
    """    
    cdef list segments = [] 
    cdef str segment
    cdef str param
    cdef int count = 0
    cdef int segment_wtype
    cdef double segment_slew = 0
    cdef double segment_start = 0
    cdef double segment_end = 1
    cdef int segment_length = wtlength
    cdef int total_segment_length = 0

    for segment in score.split(' '):
        segment_start = 0
        segment_end = 1
        segment_slew = 0
        segment_wtype = SINE
        segment_length = wtlength

        print(segment)
        match = SEGMENT_RE.match(segment)
        print(match)
        
        length = match.group('length')
        if length is not None:
            segment_length = <int>(wtlength * float(length))

        wtype = match.group('wtype')
        if wtype is not None:
            segment_wtype = wtypentf(wtype)

        start = match.group('start')
        if start is not None:
            segment_start = float(start)

        end = match.group('end')
        if end is not None:
            segment_end = float(end)

        segments += (segment_length, segment_wtype, segment_start, segment_end)
        total_segment_length += segment_length

    print(segments)

    return segments, total_segment_length

cpdef double[:] polyseg(list segments, int length):
    """ Calculate total output segment length in frames & alloc outbuf

        loop thru segments:
            - generate segment
            - write segment into outbuf

        for each segment:
            - calc the true length of the individual segment based on segment length
            - scale the size of the tmp segment to match remainder after start/end filtering
            - produce a tmp array from the waveform type (using normal wavetable or window generator)
            - copy into outbuf

        segment crossfading?

    """
    for segment_length, segment_wtype, segment_start, segment_end in segments:
        pass

    cdef double[:] out = np.zeros(length, dtype='d')

    return out


cpdef double[:] randline(int numpoints, double lowvalue=0, double highvalue=1, int wtsize=4096):
    points = np.array([ random.triangular(lowvalue, highvalue) for _ in range(numpoints) ], dtype='d')
    return interpolation._linear(points, wtsize)

cdef double[:] _window(int window_type, int length):
    cdef double[:] wt

    if window_type == RND:
        window_type = ALL_WINDOWS[rand() % LEN_WINDOWS]
        wt = _window(window_type, length)

    elif window_type == SINE:
        wt = np.sin(np.linspace(0, np.pi, length, dtype='d'))

    elif window_type == TRI:
        wt = np.abs(np.abs(np.linspace(0, 2, length, dtype='d') - 1) - 1)

    elif window_type == SAW:
        wt = np.linspace(0, 1, length, dtype='d')

    elif window_type == RSAW:
        wt = np.linspace(1, 0, length, dtype='d')

    elif window_type == HANN:
        wt = np.hanning(length)

    elif window_type == HAMM:
        wt = np.hamming(length)

    elif window_type == BART:
        wt = np.bartlett(length)

    elif window_type == BLACK:
        wt = np.blackman(length)

    elif window_type == KAISER:
        wt = np.kaiser(length, 0)

    else:
        wt = window(SINE, length)

    return wt

cdef double[:] _adsr(int framelength, int attack, int decay, double sustain, int release):
    cdef int alen = attack + decay + release
    cdef double mult = 1
    if alen > framelength:
        mult = <double>framelength / <double>alen
        attack = <int>(mult * attack)
        decay = <int>(mult * decay)
        release = <int>(mult * release)

    cdef int decay_breakpoint = decay + attack
    cdef int sustain_breakpoint = framelength - release
    cdef int decay_framelength = decay_breakpoint - attack
    cdef int release_framelength = framelength - sustain_breakpoint
    cdef double[:] out = np.zeros(framelength, dtype='d')

    for i in range(framelength):
        if i <= attack and attack > 0:
            out[i] = i / <double>attack

        elif i <= decay_breakpoint and decay_breakpoint > 0:
            out[i] = (1 - ((i - attack) / <double>decay_framelength)) * (1 - sustain) + sustain
    
        elif i <= sustain_breakpoint:
            out[i] = sustain

        else:
            out[i] = (1 - ((i - sustain_breakpoint) / <double>release_framelength)) * sustain

    return out

cpdef double[:] adsr(int length, int attack, int decay, double sustain, int release):
    return _adsr(length, attack, decay, sustain, release)

def window(int window_type, int length, double[:] data=None):
    if data is not None:
        return interpolation._linear(data, length)

    return _window(window_type, length)

cdef double[:] _wavetable(int wavetable_type, int length):
    cdef double[:] wt

    if wavetable_type == RND:
        wavetable_type = ALL_WAVETABLES[rand() % LEN_WAVETABLES]
        wt = _wavetable(wavetable_type, length)

    elif wavetable_type == SINE:
        wt = np.sin(np.linspace(-np.pi, np.pi, length, dtype='d', endpoint=False))

    elif wavetable_type == COS:
        wt = np.cos(np.linspace(-np.pi, np.pi, length, dtype='d', endpoint=False))

    elif wavetable_type == TRI:
        tmp = np.abs(np.linspace(-1, 1, length, dtype='d', endpoint=False))
        tmp = np.abs(tmp)
        wt = (tmp - tmp.mean()) * 2

    elif wavetable_type == SAW:
        wt = np.linspace(-1, 1, length, dtype='d', endpoint=False)

    elif wavetable_type == RSAW:
        wt = np.linspace(1, -1, length, dtype='d', endpoint=False)

    elif wavetable_type == SQUARE:
        tmp = np.zeros(length, dtype='d')
        duty = int(length * SQUARE_DUTY)
        tmp[:duty] = 1
        tmp[duty:] = -1
        wt = tmp

    else:
        wt = _wavetable(SINE, length)

    return wt

cpdef double[:] wavetable(int wavetable_type, int length, double[:] data=None):
    if data is not None:
        return interpolation._linear(data, length)

    return _wavetable(wavetable_type, length)

cpdef double[:] fromfile(unicode filename, int length):
    wt, _ = soundfile.read(filename, dtype='d')
    if len(wt) == length:
        return wt

    return interpolation._linear(wt, length)





from .soundbuffer cimport SoundBuffer
from . cimport wavetables

cdef class GrainCloud:
    cdef public SoundBuffer buf
    cdef public double amp
    cdef public int channels
    cdef public int samplerate
    cdef public double[:] win
    cdef public int win_length

    cdef public int[:] mask

    cdef public double freeze
    cdef public double[:] read_lfo
    cdef public int read_lfo_length
    cdef public double read_lfo_speed
    cdef public double read_jitter

    cdef public double[:] speed_lfo
    cdef public int speed_lfo_length
    cdef public double speed
    cdef public double minspeed
    cdef public double maxspeed

    cdef public double spread
    cdef public double jitter
    cdef public double density
    cdef public double mindensity
    cdef public double maxdensity
    cdef public double grains_per_sec

    cdef public double[:] density_lfo
    cdef public int density_lfo_length
    cdef public double density_lfo_speed
    cdef public double density_jitter

    cdef public double grainlength
    cdef public double[:] grainlength_lfo
    cdef public int grainlength_lfo_length
    cdef public double grainlength_lfo_speed
    cdef public double minlength
    cdef public double maxlength




import random
from PIL import Image, ImageDraw
from . import interpolation

def waveform(sound, filename=None, width=400, height=300, stroke=3, upsample_mult=5, upsample_x=20, show_axis=True):
    width *= upsample_mult
    height *= upsample_mult

    if filename is None:
        filename = 'waveform.png'

    x = range(width*upsample_x)

    img = Image.new('RGBA', (width, height), (255, 255, 255, 255))
    draw = ImageDraw.Draw(img)

    for channel in range(sound.channels):
        color = tuple([random.randint(0, 200) for _ in range(3)] + [200])
        points = [ sound[i][channel] for i in range(len(sound)) ]
        points = interpolation.linear(points, width*upsample_x)

        mapped_points = []
        for pos, point in zip(x, points):
            y = int(((point + 1) / 2) * (height - (stroke * 2)) + stroke)
            pos /= upsample_x
            mapped_points += [ (pos, y) ]
            draw.ellipse((pos-stroke, y-stroke, pos+stroke, y+stroke), fill=color)
        draw.line(mapped_points, fill=color, width=stroke)

    if show_axis:
        draw.line((0, height/2, width, height/2), fill=(0,0,0,255), width=stroke//4)

    img.thumbnail((width//upsample_mult, height//upsample_mult))
    img.save(filename)




import numbers
from . import soundbuffer
from . import tune

class Sampler:
    def __init__(self, filename, pitch=None):
        self.pitch = tune.ntf(pitch or 'a4')
        self.sound = soundbuffer.SoundBuffer(filename=filename)

    def play(self, freq=440):
        speed = freq / self.pitch
        sound = self.sound.speed(speed)
        return sound

    def init_sampledb(self):
        self.db = sqlite3.connect(':memory:')
        self.db.cursor.execute('create table samples (freq real, sample binary);')




import numbers
import numpy as np
from .soundbuffer cimport SoundBuffer
from . import wavetables
from . cimport wavetables
from . cimport interpolation

cimport cython
from cpython.array cimport array, clone

cdef inline int DEFAULT_SAMPLERATE = 44100
cdef inline int DEFAULT_CHANNELS = 2
cdef inline double MIN_PULSEWIDTH = 0.0001

cdef class Osc:
    """ 1d or 2d wavetable osc
    """
    cdef double freq
    cdef double amp

    cdef double[:] wavetable
    cdef list wavetables
    cdef double lfo_freq
    cdef object lfo

    cdef double[:] window
    cdef double pulsewidth

    cdef double[:] mod
    cdef double mod_range
    cdef double mod_freq

    cdef double phase
    cdef double win_phase
    cdef double mod_phase

    cdef int channels
    cdef int samplerate
    cdef int wtsize

    def __init__(
            self, 
            object wavetable=None, 
              list stack=None,
            double freq=440, 
            double amp=1, 
            double pulsewidth=1,
            double phase=0, 

            object window=None, 
            double win_phase=0, 

            object mod=None, 
            double mod_freq=0.2, 
            double mod_range=0, 
            double mod_phase=0, 

            object lfo=None, 
            double lfo_freq=0.2, 

            int wtsize=4096,
            int channels=DEFAULT_CHANNELS,
            int samplerate=DEFAULT_SAMPLERATE,
        ):

        self.freq = freq
        self.amp = amp
        self.phase = phase

        self.channels = channels
        self.samplerate = samplerate
        self.wtsize = wtsize

        self.pulsewidth = pulsewidth if pulsewidth >= MIN_PULSEWIDTH else MIN_PULSEWIDTH

        if isinstance(wavetable, int):
            # Create wavetable type if wavetable is an integer flag
            self.wavetable = wavetables._wavetable(wavetable, self.wtsize)

        elif isinstance(wavetable, wavetables.Wavetable):
            self.wavetable = wavetable.data

        elif isinstance(wavetable, SoundBuffer):
            if wavetable.channels > 1:
                wavetable = wavetable.remix(1)
            self.wavetable = wavetable.frames.base.flatten()

        elif wavetable is not None:
            self.wavetable = array('d', wavetable)

        if stack is not None:
            self.wavetables = []
            for i, wt in enumerate(stack):
                if isinstance(wt, int):
                    self.wavetables += [ wavetables._wavetable(wt, self.wtsize) ]
                elif isinstance(wt, wavetables.Wavetable):
                    self.wavetables += [ interpolation._linear(wt.data, self.wtsize) ]
                elif isinstance(wt, SoundBuffer):
                    if wt.channels > 1:
                        wt = wt.remix(1)
                    self.wavetables += [ interpolation._linear(wt.frames.base.flatten(), self.wtsize) ]
                else:
                    self.wavetables += [ interpolation._linear(array('d', wt), self.wtsize) ]

        self.win_phase = win_phase
        if isinstance(window, int):
            self.window = wavetables._window(window, self.wtsize)
        elif isinstance(window, wavetables.Wavetable):
            self.window = interpolation._linear(window.data, self.wtsize)
        elif isinstance(window, SoundBuffer):
            if window.channels > 1:
                window = window.remix(1)
            self.window = interpolation._linear(window.frames.base.flatten(), self.wtsize)
        elif window is not None:
            self.window = interpolation._linear(array('d', window), self.wtsize)
        else:
            self.window = None

        self.mod_range = mod_range
        self.mod_phase = mod_phase
        self.mod_freq = mod_freq
        if isinstance(mod, int):
            self.mod = wavetables._window(mod, self.wtsize)
        elif isinstance(mod, wavetables.Wavetable):
            self.mod = interpolation._linear(mod.data, self.wtsize)
        elif isinstance(mod, SoundBuffer):
            if mod.channels > 1:
                mod = mod.remix(1)
            self.mod = interpolation._linear(mod.frames.base.flatten(), self.wtsize)
        elif isinstance(mod, list):
            self.mod = interpolation._linear(array('d', mod), self.wtsize)
        elif mod is not None:
            self.mod = mod
        else:
            self.mod = None

        self.lfo_freq = lfo_freq
        if isinstance(lfo, int):
            self.lfo = wavetables._window(lfo, self.wtsize)
        elif isinstance(lfo, wavetables.Wavetable):
            self.lfo = interpolation._linear(lfo.data, self.wtsize)
        elif isinstance(lfo, SoundBuffer):
            if lfo.channels > 1:
                lfo = lfo.remix(1)
            self.lfo = interpolation._linear(lfo.frames.base.flatten(), self.wtsize)
        elif isinstance(lfo, list):
            self.lfo = interpolation._linear(array('d', lfo), self.wtsize)
        else:
            self.lfo = None

        if stack is not None and self.lfo is None:
            self.lfo = wavetables._wavetable(wavetables.RSAW, self.wtsize)

    cpdef object play(self, 
             double length, 
             double freq=-1, 
             double amp=-1, 
             double pulsewidth=-1,
             double mod_freq=-1,
             double mod_range=-1,
        ):

        framelength = <int>(length * self.samplerate)

        if freq > 0:
            self.freq = freq

        if amp >= 0:
            self.amp = amp 

        if pulsewidth > 0:
            self.pulsewidth = pulsewidth 

        if mod_freq > 0:
            self.mod_freq = mod_freq 

        if mod_range >= 0:
            self.mod_range = mod_range 

        if self.wavetables is not None:
            return self._play2d(framelength)
        else:
            return self._play(framelength)

    cdef object _play(self, int length):
        cdef double[:,:] out = np.zeros((length, self.channels), dtype='d')

        cdef int i = 0
        cdef int wtindex = 0
        cdef int modindex = 0
        cdef int wtlength = len(self.wavetable)
        cdef int wtboundry = wtlength - 1
        cdef int modlength = 1 if self.mod is None else len(self.mod)
        cdef int modboundry = modlength - 1
        cdef double val, val_mod, nextval, nextval_mod, frac, frac_mod
        cdef int silence_length

        cdef double[:] wavetable = self.wavetable
        cdef double[:] window

        cdef unsigned int j = 0

        if self.window is not None:
            window = interpolation._linear(self.window, wtlength)
            for j in range(wtlength):
                wavetable[j] *= window[j]

        if self.pulsewidth < 1:
            silence_length = <int>((wtlength / self.pulsewidth) - wtlength)
            wavetable = np.concatenate((wavetable, np.zeros(silence_length, dtype='d')))

        cdef double isamplerate = 1.0 / self.samplerate

        for i in range(length):
            wtindex = <int>self.phase % wtlength
            modindex = <int>self.mod_phase % modlength

            val = wavetable[wtindex]
            val_mod = 1

            if wtindex < wtboundry:
                nextval = wavetable[wtindex + 1]
            else:
                nextval = wavetable[0]

            if self.mod is not None:
                if modindex < modboundry:
                    nextval_mod = self.mod[modindex + 1]
                else:
                    nextval_mod = self.mod[0]

                frac_mod = self.mod_phase - <int>self.mod_phase
                val_mod = self.mod[modindex]
                val_mod = (1.0 - frac_mod) * val_mod + frac_mod * nextval_mod
                val_mod = 1.0 + (val_mod * self.mod_range)
                self.mod_phase += self.mod_freq * modlength * isamplerate

            frac = self.phase - <int>self.phase
            val = ((1.0 - frac) * val + frac * nextval) * self.amp
            self.phase += self.freq * val_mod * wtlength * isamplerate

            for channel in range(self.channels):
                out[i][channel] = val

        return SoundBuffer(out, channels=self.channels, samplerate=self.samplerate)

    cdef object _play2d(self, int length):
        cdef double[:,:] out = np.zeros((length, self.channels), dtype='d')
        cdef double[:,:] stack = np.column_stack(self.wavetables)
        cdef int stack_depth = len(self.wavetables)
        cdef double isamplerate = (1.0 / self.samplerate)
        cdef double wt_lfo_phase_inc = len(self.lfo) * (1.0 / length)
        cdef double phase_inc = self.freq * self.wtsize * isamplerate
        cdef double wt_phase_inc = self.freq * self.wtsize * isamplerate

        cdef double wt_lfo_phase, wt_lfo_frac, wt_lfo_y0, wt_lfo_y1, wt_lfo_pos
        cdef double stack_frac, stack_phase
        cdef double phase, frac, y0, y1, val
        cdef double wt_mod_val = 1
        cdef double wt_mod_phase = 0
        cdef double wt_mod_frac = 0
        cdef double wt_mod_next = 0

        cdef int i, wt_lfo_x, stack_x, channel, wt_mod_i
        cdef int wt_mod_length = 1 if self.mod is None else len(self.mod)
        cdef int wt_mod_boundry = wt_mod_length - 1
        cdef double wt_phase = 0
        cdef double wt_mod_phase_inc = self.mod_freq * (1.0 / length) * wt_mod_length

        for i in range(length):
            if self.mod is not None:
                wt_mod_i = <int>self.mod_phase % wt_mod_length
                if wt_mod_i < wt_mod_boundry:
                    wt_mod_next = self.mod[wt_mod_i + 1]
                else:
                    wt_mod_next = self.mod[0]

                wt_mod_frac = self.mod_phase - <int>self.mod_phase
                wt_mod_val = self.mod[wt_mod_i]
                wt_mod_val = (1.0 - wt_mod_frac) * wt_mod_val + (wt_mod_frac * wt_mod_next)
                wt_mod_val = 1.0 + (wt_mod_val * self.mod_range)
                self.mod_phase += wt_mod_phase_inc

            # Calculate stack LFO position
            wt_lfo_phase = i * wt_lfo_phase_inc
            wt_lfo_x = <int>wt_lfo_phase
            wt_lfo_frac = wt_lfo_phase - wt_lfo_x
            wt_lfo_y0 = self.lfo[wt_lfo_x % len(self.lfo)]
            wt_lfo_y1 = self.lfo[(wt_lfo_x + 1) % len(self.lfo)]
            wt_lfo_pos = (1.0 - wt_lfo_frac) * wt_lfo_y0 + (wt_lfo_frac * wt_lfo_y1)

            # Calculate wavetable value based on LFO position
            stack_phase = wt_lfo_pos * stack_depth
            stack_x = <int>stack_phase
            wt_phase += (wt_phase_inc * wt_mod_val)
            wt_x = <int>wt_phase
            stack_frac = stack_phase - stack_x
            y0 = self.wavetables[stack_x % stack_depth][wt_x % self.wtsize]
            y1 = self.wavetables[(stack_x + 1) % stack_depth][wt_x % self.wtsize]
            val = (1.0 - stack_frac) * y0 + (stack_frac * y1)

            for channel in range(self.channels):
                out[i][channel] = val * self.amp

        return SoundBuffer(out, channels=self.channels, samplerate=self.samplerate)


cdef class Fold:
    """ Folding wavetable oscilator
    """
    cdef public double[:] wavetable
    cdef public double[:] factors
    cdef public double factFreq
    cdef public double freq
    cdef public double amp
    cdef int channels
    cdef int samplerate
    cdef int wtsize


    def __init__(
            self, 
            double[:] wavetable=None, 
            double[:] factors=None, 
            double freq=440, 
            double factFreq=1,
            double amp=1, 
            int wtsize=4096,
            int channels=DEFAULT_CHANNELS,
            int samplerate=DEFAULT_SAMPLERATE,
        ):

        self.freq = freq
        self.factFreq = factFreq
        self.amp = amp

        self.channels = channels
        self.samplerate = samplerate
        self.wtsize = wtsize

        self.wavetable = wavetable
        self.factors = factors

    def play(self, double length):
        return self._play(length)

    cdef SoundBuffer _play(self, double length):
        cdef int framelength = <int>(length * self.samplerate)
        cdef int i
        cdef int channel = 0

        cdef double valWaveform, valNextWaveform, freq, factFreq
        cdef int cIndexWaveform = 0
        cdef double fracWaveform = 0

        cdef double valFactors, valNextFactors
        cdef int cIndexFactors = 0
        cdef double fracFactors = 0

        cdef double indexWaveform = 0
        cdef double indexFactors = 0

        cdef int lenWaveform = len(self.wavetable)
        cdef int lenFactors = len(self.factors)

        cdef double fold_out = 0
        cdef double last_value = 0 
        cdef int pos_thresh = 1 
        cdef int neg_thresh = -1 
        cdef int state = 1
        cdef double difference = 0

        cdef double[:,:] out = np.zeros((framelength, self.channels), dtype='d')

        for i in range(framelength):
            # Interp waveform wavetable
            cIndexWaveform = <int>indexWaveform % (lenWaveform - 1)
            valWaveform = self.wavetable[cIndexWaveform]
            valNextWaveform = self.wavetable[cIndexWaveform + 1]
            fracWaveform = indexWaveform - <int>indexWaveform
            valWaveform = (1.0 - fracWaveform) * valWaveform + fracWaveform * valNextWaveform

            # Interp factors wavetable
            cIndexFactors = <int>indexFactors % (lenFactors - 1)
            valFactors = self.factors[cIndexFactors]
            valNextFactors = self.factors[cIndexFactors + 1]
            fracFactors = indexFactors - <int>indexFactors
            valFactors = (1.0 - fracFactors) * valFactors + fracFactors * valNextFactors
            valWaveform *= valFactors

            # Do the folding 
            # the magic is ripped from guest's posts on muffwiggler: 
            # http://www.muffwiggler.com/forum/viewtopic.php?p=1586526#1586526
            difference = valWaveform - last_value
            last_value = valWaveform
            if state == 1:
                fold_out -= difference
            else:
                fold_out += difference

            if fold_out >= pos_thresh:
                state ^= 1
                fold_out = pos_thresh - (fold_out - pos_thresh)
            elif fold_out <= neg_thresh:
                state ^= 1
                fold_out = neg_thresh - (fold_out - neg_thresh)

            for channel in range(self.channels):
                out[i][channel] = self.amp * fold_out

            indexWaveform += self.freq * lenWaveform * (1.0 / self.samplerate)
            indexFactors += self.factFreq * lenFactors * (1.0 / self.samplerate)

        return SoundBuffer(out, channels=self.channels, samplerate=self.samplerate)




import numbers
import random
import reprlib

import soundfile
import numpy as np
cimport cython
from libc cimport math

from . cimport wavetables
from . cimport interpolation
from . cimport grains
from .defaults cimport DEFAULT_SAMPLERATE, DEFAULT_CHANNELS, DEFAULT_SOUNDFILE

cdef double[:,:] _remix(double[:,:] values, int framelength, int channels):
    if len(values) == framelength and values.shape[1] == channels:
        return values

    cdef double[:,:] out = None
    cdef int i = 0
    cdef int channel = 0

    if values.shape[1] == channels:
        for channel in range(channels):
            out[:,channel] = interpolation._linear(values[:,channel], framelength)
        return out

    elif channels <= 1:
        return np.sum(values, 1).reshape(len(values), 1)

    else:
        out = np.zeros((framelength, channels))
        for i in range(framelength):
            out[i] = interpolation._linear(values[i], channels)
        return out

cdef double[:,:] _mul2d(double[:,:] output, double[:,:] values):
    return np.multiply(output, _remix(values, len(output), output.shape[1]))

cdef double[:,:] _mul1d(double[:,:] output, double[:] values):
    cdef int channel = 0
    cdef int channels = output.shape[1]
    cdef int framelength = len(output)

    if len(values) != framelength:
        values = interpolation._linear(values, framelength)

    for channel in range(channels):
        output.base[:,channel] *= values

    return output

cdef class SoundBuffer:
    """ A sequence of audio frames 
        representing a buffer of sound.
    """

    def __cinit__(self, 
            object frames=None, 
            double length=-1, 
               int channels=-1, 
               int samplerate=DEFAULT_SAMPLERATE, 
           unicode filename=None, 
            double start=0, 
       double[:,:] buf=None):
        """ Data may be loaded into the SoundBuffer from these sources as keyword arguments, 
            in decending order of precedence:

              - `filename`: a unicode string relative path or a full path to load from a soundfile
              - `buf`: a memoryview compatible 2D buffer of doubles <double[:,:]>
              - `frames`: a 1D or 2D python iterable (eg a list) or a 1D numpy array/buffer

            Data loaded via the `filename` keyword can load any soundfile type supported by libsndfile,
            including old favorites like WAV, AIFF, FLAC, and OGG. But no, not MP3. (At least not yet: https://github.com/erikd/libsndfile/issues/258)
            If the `length` param is given, only that amount will be read from the soundfile.
            Use the `start` param to specify a position (seconds) in the soundfile to begin reading from.
            No samplerate conversion is done (yet) so be careful to match the samplerate of the start param with that 
            of the soundfile being read. (Or don't!)
            Overflows (reading beyond the boundries of the file) are filled with silence so the length 
            param is always respected and will return a SoundBuffer of the requested length. This is 
            just a wrapper around pysndfile's soundfile.read method.

            Data loaded via the `buf` keyword must be interpretable by cython as a memoryview on a 2D array 
            of doubles. (A numpy ndarray for example.)

            Single channel (1D) data loaded via the `frames` keyword will be copied into the given number of channels
            specified by the `channels` param. (And if no value is given, a mono SoundBuffer will be created.) 
            Handy for doing synthesis prototyping with python lists.

            Multichannel (2D) data loaded via the `frames` keyword will ignore the `channels` 
            param and adapt to what was given by attempting to read the data into a numpy array
            and take the number of channels from its shape.
            
            The remaining sources (`filename`, `buf`) will use the number of channels found in the 
            data and ignore the value passed to the SoundBuffer initializer.

            If none of those sources are present, but a `length` param has been given, 
            the SoundBuffer will be initialized with silence to the given length in seconds.
            
            Otherwise an empty SoundBuffer is created with a length of zero.
        """
        self.samplerate = samplerate
        self.channels = channels
        cdef int framestart = 0
        cdef int framelength = <int>(length * self.samplerate)

        if filename is not None:
            framestart = <int>(start * self.samplerate)
            self.frames, self.samplerate = soundfile.read(filename, framelength, framestart, dtype='float64', fill_value=0)
            self.channels = self.frames.shape[1]

        elif buf is not None:
            self.frames = buf
            self.channels = self.frames.shape[1]

        elif frames is not None:
            if isinstance(frames, list) or len(frames.shape) == 1:
                if self.channels <= 0:
                    self.channels = 1
                self.frames = np.column_stack([ frames for _ in range(self.channels) ])
            else:
                self.frames = np.array(frames, dtype='d')
                self.channels = self.frames.shape[1]

        elif length > 0:
            if self.channels <= 0:
                self.channels = DEFAULT_CHANNELS
            self.frames = np.zeros((framelength, self.channels))
        else:
            self.frames = None


    #############################################
    # (+) Addition & concatenation operator (+) #
    #############################################
    def __add__(self, value):
        """ Add a number to every sample in this SoundBuffer or concatenate with another 
            SoundBuffer or iterable with compatible dimensions
        """
        if self.frames is None:
            if isinstance(value, numbers.Real):
                return SoundBuffer(channels=self.channels, samplerate=self.samplerate)
            elif isinstance(value, SoundBuffer):
                return value.copy()
            else:
                return SoundBuffer(value, channels=self.channels, samplerate=self.samplerate)

        cdef int length = len(self.frames)
        cdef double[:,:] out = np.zeros((length, self.channels))

        if isinstance(value, numbers.Real):
            out = np.add(self.frames, value)
        elif isinstance(value, SoundBuffer):
            if value.frames is None:
                out = self.frames.copy()
            else:
                out = np.vstack((self.frames, value.frames))
        else:
            try:
                out = np.vstack((self.frames, value))
            except TypeError as e:
                return NotImplemented

        return SoundBuffer(out, channels=self.channels, samplerate=self.samplerate)

    def __iadd__(self, value):
        """ In place add either adding number to every value without copy, or 
            directly extending internal frame buffer.
        """
        if self.frames is None:
            if isinstance(value, numbers.Real):
                return self
            elif isinstance(value, SoundBuffer):
                return value.copy()
            else:
                return SoundBuffer(value, channels=self.channels, samplerate=self.samplerate)

        if isinstance(value, numbers.Real):
            self.frames = np.add(self.frames, value)
        else:
            try:
                self.frames = np.vstack((self.frames, value))
            except TypeError as e:
                return NotImplemented

        return self

    def __radd__(self, value):
        return self + value


    cdef SoundBuffer _adsr(self, double attack, double decay, double sustain, double release):
        sustain = min(max(0, sustain), 1)
        attack = max(0, attack)
        decay = max(0, decay)
        release = max(0, release)

        cdef double alen = attack + decay + release
        cdef double mult = 1
        if alen > self.dur:
            mult = self.dur / alen
            attack *= mult
            decay *= mult
            release *= mult

        cdef int framelength = len(self)
        cdef int attack_breakpoint = <int>(<double>self.samplerate * attack)
        cdef int decay_breakpoint = <int>(<double>self.samplerate * decay) + attack_breakpoint
        cdef int sustain_breakpoint = framelength - <int>(release * <double>self.samplerate)
        cdef int decay_framelength = decay_breakpoint - attack_breakpoint
        cdef int release_framelength = framelength - sustain_breakpoint

        for i in range(framelength):
            if i <= attack_breakpoint and attack_breakpoint > 0:
                self.frames.base[i] = self.frames.base[i] * (i / <double>attack_breakpoint)

            elif i <= decay_breakpoint and decay_breakpoint > 0:
                self.frames.base[i] = self.frames.base[i] * ((1 - ((i - attack_breakpoint) / <double>decay_framelength)) * (1 - sustain) + sustain)
        
            elif i <= sustain_breakpoint:
                self.frames.base[i] = self.frames.base[i] * sustain

            else:
                release = 1 - ((i - sustain_breakpoint) / <double>release_framelength)
                self.frames.base[i] = self.frames.base[i] * release * sustain

        return self

    def adsr(self, double a=0, double d=0, double s=1, double r=0):
        return self._adsr(a, d, s, r)


    ########################
    # (&) Mix operator (&) #
    ########################
    def __and__(self, value):
        """ Mix two SoundBuffers or two 2d arrays with compatible dimensions

            >>> sound = SoundBuffer(length=1, channels=2, samplerate=44100)
            >>> sound2 = SoundBuffer(length=0.35, channels=2, samplerate=44100)
            >>> out = sound2 & sound
            >>> len(out)
            1.0
        """

        cdef double[:,:] out

        try:
            out = np.add(self.frames, value[:len(self.frames)])
            return SoundBuffer(out, channels=self.channels, samplerate=self.samplerate)
        except TypeError as e:
            return NotImplemented

    def __iand__(self, value):
        """ Mix in place two SoundBuffers or two 2d arrays with compatible dimensions
        """
        self.frames = np.add(self.frames, value[:len(self.frames)])
        return self

    def __rand__(self, value):
        return self & value


    ##############
    # Truthiness #
    ##############
    def __bool__(self):
        return bool(len(self))


    #############################
    # ([:]) Frame slicing ([:]) #
    #############################
    def __getitem__(self, position):
        """ Slice into this SoundBuffer, returning a copy of a portion of the buffer, 
            or a tuple of samples representing a single frame.

            >>> sound = SoundBuffer(length=1, samplerate=44100, channels=2)
            >>> bit = sound[0:4410]
            >>> len(bit)
            0.1
            >>> sound[0]
            (0,0)
        """
        if isinstance(position, int):
            return tuple(self.frames[<int>position])

        cdef double[:, :] out

        if not position.stop:
            out = self.frames[position.start:]

        elif not position.start:
            out = self.frames[:position.stop]

        else:
            out = self.frames[position.start:position.stop]

        return SoundBuffer(out, self.channels, self.samplerate)


    ###############
    # Lengthiness #
    ###############
    def __len__(self):
        """ Return the length of this SoundBuffer in frames.
            >>> sound = SoundBuffer(length=1, samplerate=44100, channels=2)
            >>> len(sound)
            44100
        """
        return 0 if self.frames is None else len(self.frames)

    @property
    def dur(self):
        """ Duration of the SoundBuffer in seconds
        """
        return len(self) / <double>self.samplerate


    ###################################
    # (*) Multiplication operator (*) #
    ###################################
    def __mul__(self, value):
        """ Multiply a SoundBuffer by a number, iterable or SoundBuffer and return a new SoundBuffer

            >>> sound = SoundBuffer(length=1, channels=2, samplerate=44100)

            # Multiply every sample in `sound` by 0.5
            >>> sound *= 0.5

            # Multiply every sample by an interpolated copy of `wavetable` 
            # `wavetable` can be a 1d flat array/list/iterable, or a SoundBuffer
            # This will modulate the amplitude of the SoundBuffer with the wavetable.
            >>> wavetable = [0, 0.25, 0.5, 0.25, 0.75, 1]
            >>> sound = sound * wavetable

            >>> sound2 = SoundBuffer(length=0.5, channels=2, samplerate=44100)
            >>> sound = sound * sound2

        """
        cdef int length = len(self.frames)
        cdef double[:,:] out
        cdef double[:] wavetable

        if isinstance(value, numbers.Real):
            out = np.multiply(self.frames, value)

        elif isinstance(value, SoundBuffer):
            out = _mul2d(self.frames, value.frames)

        elif isinstance(value, list):
            out = _mul1d(self.frames, np.array(value))

        else:
            try:
                if len(value.shape) == 1:
                    out = _mul1d(self.frames, value)
                else:
                    out = _mul2d(self.frames, value)
            except TypeError:
                return NotImplemented

        return SoundBuffer(out, channels=self.channels, samplerate=self.samplerate)

    def __imul__(self, value):
        """ Multiply a SoundBuffer in place by a number, iterable or SoundBuffer
        """
        if isinstance(value, numbers.Real):
            self.frames = np.multiply(self.frames, value)

        elif isinstance(value, SoundBuffer):
            self.frames = _mul2d(self.frames, value.frames)

        elif isinstance(value, list):
            self.frames = _mul1d(self.frames, np.array(value))

        else:
            try:
                if len(value.shape) == 1:
                    self.frames = _mul1d(self.frames, value)
                else:
                    self.frames = _mul2d(self.frames, value)
            except TypeError:
                return NotImplemented

        return self

    def __rmul__(self, value):
        return self * value


    ################################
    # (-) Subtraction operator (-) #
    ################################
    def __sub__(self, value):
        """ Subtract a number from every sample in this SoundBuffer or  
            from another SoundBuffer or iterable with compatible dimensions
        """
        cdef double[:,:] out

        if isinstance(value, numbers.Real):
            out = np.subtract(self.frames, value)
        if isinstance(value, SoundBuffer):
            out = np.subtract(self.frames, value.frames)
        else:
            try:
                out = np.subtract(self.frames, value[:,None])
            except TypeError as e:
                return NotImplemented

        return SoundBuffer(out, channels=self.channels, samplerate=self.samplerate)

    def __isub__(self, value):
        """ In place subtract a number from every sample in this SoundBuffer or  
            from another SoundBuffer or iterable with compatible dimensions
        """
        if isinstance(value, numbers.Real):
            self.frames = np.subtract(self.frames, value)
        if isinstance(value, SoundBuffer):
            self.frames = np.subtract(self.frames, value.frames)
        else:
            try:
                self.frames = np.subtract(self.frames, value.frames)
            except TypeError as e:
                return NotImplemented

    def __rsub__(self, value):
        return self - value


    ##############################
    # Pickling & stringification #
    ##############################
    def __reduce__(self):
        """ Tells the pickle module how to rebuild a pickled instance of a SoundBuffer
        """
        return (rebuild_buffer, (np.asarray(self.frames), self.channels, self.samplerate))

    def __repr__(self):
        """ Return a string representation of this SoundBuffer
        """
        return 'SoundBuffer(samplerate={}, channels={}, frames={})'.format(self.samplerate, self.channels, self.frames)


    ##########################
    # Transformation methods #
    ##########################
    def clear(self, int length=-1):
        """ Writes zeros into the buffer, adjusting the
            size if length > 0.
        """
        if length <= 0:
            self.frames = None
        else:
            self.frames = np.zeros((length, self.channels))

    def cloud(self, double length=-1, *args, **kwargs):
        """ Create a new GrainCloud from this SoundBuffer
        """
        if length <= 0:
            length = self.dur
        return grains.GrainCloud(self, *args, **kwargs).play(length)

    def copy(self):
        """ Return a new copy of this SoundBuffer.
        """
        if self.frames is None:
            return SoundBuffer(channels=self.channels, samplerate=self.samplerate)
        return SoundBuffer(np.array(self.frames, copy=True), self.channels, self.samplerate)

    def clip(self, minval=-1, maxval=1):
        return SoundBuffer(np.clip(self.frames, minval, maxval), self.channels, self.samplerate)
        
    def cut(self, double start=0, double length=1):
        """ Copy a portion of this soundbuffer, returning 
            a new soundbuffer with the selected slice.

            This is called `cut` for historical reasons, but it 
            might be better understood if renamed to `slice` as it 
            doesn't modify the source SoundBuffer in any way.
            
            The `start` param is a position in seconds to begin 
            cutting, and the `length` param is the cut length in seconds.

            Overflowing values that exceed the boundries of the source SoundBuffer 
            will return a SoundBuffer padded with silence so that the `length` param 
            is always respected.
        """
        cdef int readstart = <int>(start * self.samplerate)
        cdef int outlength = <int>(length * self.samplerate)
        cdef int readend = outlength + readstart
        cdef int readlength = len(self) - readstart

        cdef double[:,:] out = np.zeros((outlength, self.channels))
        if readlength >= 0:
            out[0:readlength] = self.frames[readstart:readend]

        return SoundBuffer(out, self.channels, self.samplerate)

    def rcut(self, double length=1):
        """ Copy a portion of this SoundBuffer of the 
            given length in seconds starting from a random 
            position within it. 
            
            This will always return a complete SoundBuffer 
            without overflows or added silence, and the entire 
            sound will be returned without added silence if a length 
            that exceeds the length of the source SoundBuffer is given -- 
            unlike SoundBuffer.cut() which will pad results with silence 
            to preserve the length param if an invalid or overflowing offset 
            position is given.
        """
        cdef double maxlen = self.dur - length
        if maxlen <= 0:
            return self
        cdef double start = random.triangular(0, maxlen)
        return self.cut(start, length)

    cdef void _dub(self, SoundBuffer sound, int framepos=0):
        cdef int dubsnd_length = len(sound.frames)
        cdef int source_length = len(self.frames)
        cdef int total_length = framepos + dubsnd_length
        if self.frames is None or source_length == 0:
            self.frames = np.zeros((total_length, self.channels), dtype='float64')

        elif source_length < total_length:
            self.frames = np.vstack((
                    self.frames.copy(), 
                    np.zeros((total_length - source_length, self.channels), dtype='float64')
                ))

        #cdef int i = 0
        #cdef int c = 0
        #for i in range(dubsnd_length):
        #    for c in range(self.channels):
        #        self.frames[framepos+i][c] += sound[i][c]

        self.frames.base[framepos:total_length] += sound.frames

    def dub(self, sounds, double pos=-1, int framepos=0):
        """ Dub a sound or iterable of sounds into this soundbuffer
            starting at the given position in fractional seconds.
            >>> snd.dub(snd2, 3.2)

            To dub starting at a specific frame position use:
            >>> snd.dub(snd3, framepos=111)
        """
        cdef int numsounds
        cdef int sound_index

        if pos >= 0:
            framepos = <int>(pos * self.samplerate)

        if isinstance(sounds, SoundBuffer):
            self._dub(sounds, framepos)
        else:
            numsounds = len(sounds)
            try:
                for sound_index in range(numsounds):
                    self._dub(sounds[sound_index], framepos)

            except TypeError as e:
                raise TypeError('Please provide a SoundBuffer or list of SoundBuffers for dubbing') from e

    def env(self, int window_type=-1):
        """ Apply an amplitude envelope 
            to the sound of the given type.

            To modulate a sound with an arbitrary 
            iterable, simply do:

            >>> snd * iterable

            Where iterable is a list, array, or SoundBuffer with 
            the same # of channels and of any length
        """
        cdef int length = len(self)
        cdef double[:] wavetable = None
        wavetable = wavetables._window(window_type, length)

        if wavetable is None:
            raise TypeError('Invalid envelope type')

        return self * wavetable

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void _fill(self, double[:,:] frames):
        """ Fill current w/frames
        """
        cdef int frame_index, channel_index = 0
        cdef int framelen = len(frames)
        cdef int length = len(self.frames)
        cdef int minlen = min(framelen, length)

        for frame_index in range(minlen):
            for channel_index in range(self.channels):
                self.frames[frame_index][channel_index] = frames[frame_index][channel_index]

    def fill(self, int length):
        """ Truncate the buffer to the given length or 
            loop the contents of the buffer up to the 
            given length in frames.
        """
        cdef int mult = 0 if length <= 0 or len(self) == 0 else length // len(self)
        out = SoundBuffer(frames=self.frames[:,:], channels=self.channels, samplerate=self.samplerate)

        if mult > 1:
            out = out.repeat(mult)

        return out[:length]

    def grains(self, double minlength, double maxlength=-1):
        """ Iterate over the buffer in fixed-size grains.
            If a second length is given, iterate in randomly-sized 
            grains, given the minimum and maximum sizes.
        """
        cdef int framesread = 0
        cdef int maxframes = <int>(maxlength * self.samplerate)
        cdef int minframes = <int>(minlength * self.samplerate)
        cdef int grainlength = minframes
        cdef int sourcelength = len(self)
        while framesread < sourcelength:
            if minlength > 0 and maxlength > 0:
                grainlength = random.randint(minframes, maxframes)

            yield self[framesread:framesread+grainlength]

            framesread += grainlength

    def max(self):
        return np.amax(self.frames)
 
    def mix(self, sounds):
        """ Mix this sound in place with a sound or iterable of sounds
        """
        cdef SoundBuffer sound
        cdef int numsounds
        cdef int sound_index
        if isinstance(sounds, SoundBuffer):
            self &= sounds
        else:
            numsounds = len(sounds)
            try:
                for sound_index in range(numsounds):
                    self &= sounds[sound_index] 
            except TypeError as e:
                raise TypeError('Please provide a SoundBuffer or list of SoundBuffers for mixing') from e

    def pad(self, double start=0, double end=0):
        """ Pad this sound with silence at start or end
        """
        if start <= 0 and end <= 0: 
            return self

        cdef int framestart = <int>(start * self.samplerate)
        cdef int frameend = <int>(end * self.samplerate)

        cdef double[:,:] out

        if framestart > 0:
            out = np.vstack((np.zeros((framestart, self.channels)), self.frames))

        if frameend > 0:
            out = np.vstack((self.frames, np.zeros((frameend, self.channels))))
 
        return SoundBuffer(out, channels=self.channels, samplerate=self.samplerate)

    def pan(self, double pos=0.5, unicode method=None, int start=0):
        """ Pan a stereo sound from pos=0 (hard left) 
            to pos=1 (hard right)

            Different panning strategies can be chosen 
            by passing a value to the `method` param.

            method='constant'
                Constant (square) power panning.
                This is the default.

            method='linear'
                Simple linear panning.

            method='sine'
                Variation on constant power panning 
                using sin() and cos() to shape the pan.
                Taken from the floss manuals csound manual.

            method='gogins'
                Also taken from the csound manual -- 
                Michael Gogins' variation on the above 
                which uses a different part of the sinewave.
        """
        if method is None:
            method = u'constant'

        cdef double[:,:] out = self.frames
        cdef int length = len(self)
        cdef int channel = 0

        if method == 'constant':
            for channel in range(self.channels):
                if channel % 2 == 0:
                    out.base[:,channel] *= math.sqrt(pos)
                else:
                    out.base[:,channel] *= math.sqrt(1 - pos)

        elif method == 'linear':
            for channel in range(self.channels):
                if channel % 2 == 0:
                    out.base[:,channel] *= pos
                else:
                    out.base[:,channel] *= 1 - pos

        elif method == 'sine':
            for channel in range(self.channels):
                if channel % 2 == 0:
                    out.base[:,channel] *= math.sin(pos * (math.pi / 2))
                else:
                    out.base[:,channel] *= math.cos(pos * (math.pi / 2))

        elif method == 'gogins':
            for channel in range(self.channels):
                if channel % 2 == 0:
                    out.base[:,channel] *= math.sin((pos + 0.5) * (math.pi / 2))
                else:
                    out.base[:,channel] *= math.cos((pos + 0.5) * (math.pi / 2))

        return SoundBuffer(out, channels=self.channels, samplerate=self.samplerate)

    def remix(self, int channels):
        return SoundBuffer(_remix(self.frames, len(self.frames), channels), self.samplerate, channels)

    def repeat(self, int reps=2):
        if reps <= 1:
            return SoundBuffer(self.frames)

        out = SoundBuffer(self.frames)
        for _ in range(reps-1):
            out += self

        return out

    def reverse(self):
        self.frames = np.flip(self.frames, 0)

    def reversed(self):
        return SoundBuffer(np.flip(self.frames, 0), channels=self.channels, samplerate=self.samplerate)

    cdef double[:,:] _speed(self, double speed):
        speed = speed if speed > 0 else 0.001
        cdef int length = <int>(len(self) * (1.0 / speed))
        cdef double[:,:] out = np.zeros((length, self.channels))
        cdef int channel

        for channel in range(self.channels):
            out[:,channel] = interpolation._linear(np.asarray(self.frames[:,channel]), length)

        return out

    def speed(self, double speed):
        """ Change the speed of the sound
        """
        return SoundBuffer(self._speed(speed), channels=self.channels, samplerate=self.samplerate)

    def stretch(self, double length, double grainlength=60):
        """ Change the length of the sound without changing the pitch.
            Granular-only implementation at the moment.
        """
        return grains.GrainCloud(self, grainlength=grainlength).play(length)

    def taper(self, double length):
        cdef int framelength = <int>(self.samplerate * length)
        return self * wavetables._adsr(len(self), framelength, 0, 1, framelength)

    def transpose(self, double speed, double grainlength=60):
        """ Change the pitch of the sound without changing the length.
            This is just a wrapper for `GrainCloud` with `speed` and `grainlength` 
            settings exposed, returning a cloud equal in length to the source sound.

            TODO accept: from/to hz, notes, midi notes, intervals
        """
        return grains.GrainCloud(self, 
                speed=speed, 
                grainlength=grainlength, 
            ).play(self.dur)

    def write(self, unicode filename=None):
        """ Write the contents of this buffer to disk 
            in the given audio file format. (WAV, AIFF, AU)
        """
        return soundfile.write(filename, self.frames, self.samplerate)


cdef class RingBuffer:
    """ Simple fixed-length ring buffer, lacking 
        the transformations and operator 
        overloading found in SoundBuffer but 
        useful for loopers, delay lines, etc.
        Also, this always speaks in frame lengths rather 
        than time units expressed in seconds.
        It's mostly just for the astrid input sampler right now.
    """
    def __cinit__(self, int length, int channels=DEFAULT_CHANNELS, int samplerate=DEFAULT_SAMPLERATE, double[:,:] frames=None):
        self.samplerate = samplerate
        self.channels = channels
        self.length = length

        if frames is not None:
            self.frames = frames
        else:
            self.frames = np.zeros((length, channels))

        self.copyout = np.zeros((length, channels))
        self.write_head = 0

    def __reduce__(self):
        return (rebuild_ringbuffer, (self.length, self.channels, self.samplerate, self.frames))

    def __len__(self):
        return self.length

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef double[:,:] _read(self, int length, int offset):
        """ Read a length of frames from the buffer
            where length is at maximum the length of 
            the buffer, always returning most recently 
            written frames.
        """
        cdef int i = 0
        cdef int read_head = 0

        if length > self.length:
            length = self.length

        read_head = (self.write_head - length + offset) % self.length
        for i in range(length):
            self.copyout[i] = self.frames[read_head]
            read_head = (read_head + 1) % self.length

        return self.copyout[:length]

    def read(self, int length, int offset=0):
        return self._read(length, offset)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void _write(self, double[:,:] frames):
        """ Write given frames into the buffer, advancing 
            the write head and wrapping as needed.
        """
        cdef int i = 0
        cdef int inputlength = len(frames)
        
        # write forward in the buffer, from the current
        # write head position, until all frames are written
        # wrapping the write head at the boundry of the buffer
        for i in range(inputlength):
            self.frames[self.write_head] = frames[i]
            self.write_head = (self.write_head + 1) % self.length

    def write(self, double[:,:] frames):
        self._write(frames)


cpdef object rebuild_buffer(double[:,:] frames, int channels, int samplerate):
    return SoundBuffer(frames, channels=channels, samplerate=samplerate)

cpdef object rebuild_ringbuffer(int length, int channels, int samplerate, double[:,:] frames):
    return RingBuffer(length, channels=channels, samplerate=samplerate, frames=frames)





import re
import math
import random

a0 = 27.5 
a4 = a0 * 2**4
default_key = 'c'
match_roman = '[ivIV]?[ivIV]?[iI]?'

intervals = {
    'P1': 0, 
    'm2': 1, 
    'M2': 2, 
    'm3': 3, 
    'M3': 4, 
    'P4': 5, 
    'TT': 6, 
    'P5': 7, 
    'm6': 8, 
    'M6': 9, 
    'm7': 10, 
    'M7': 11,
    'P8': 12,
    'm9': 13,
    'M9': 14,
    'm10': 15,
    'M10': 16,
    'P11': 17,
    'd12': 18, # TT + P8
    'P12': 19,
    'm13': 20,
    'M13': 21,
    'm14': 22,
    'M14': 23,
    'P15': 24
}

# what's the best way to handle inversions?
base_qualities = {
    '^': ['P1', 'M3', 'P5'],   # major
    '-': ['P1', 'm3', 'P5'],   # minor
    '*': ['P1', 'm3', 'TT'],   # diminished 
    '+': ['P1', 'M3', 'm6'],   # augmented
}

extensions = {
    '7': ['m7'],               # dominant 7th
    '^7': ['M7'],              # major 7th
    '9': ['m7', 'M9'],         # dominant 9th
    '^9': ['M7', 'M9'],        # major 9th
    '11': ['m7', 'M9', 'P11'], # dominant 11th
    '^11': ['M7', 'M9', 'P11'],
    '69': ['M6', 'M9'],        # dominant 6/9
    '6': ['M6'],        
}

chord_romans = {
    'i': 0,
    'ii': 2,
    'iii': 4,
    'iv': 5,
    'v': 7,
    'vi': 9,
    'vii': 11,
}

# Common root movements
progressions = {
    'I': ['iii', 'vi', 'ii', 'IV', 'V', 'vii*'],
    'i': ['VII', 'III', 'VI', 'ii*', 'iv', 'V', 'vii*'],
    'ii': ['V', 'vii*'],
    'iii': ['vi'],
    'III': ['VI'],
    'IV': ['V', 'vii*'],
    'iv': ['V', 'vii*'],
    'V': ['I', 'i'], # a pivot
    'vi': ['ii', 'IV'],
    'VI': ['ii*', 'iv'],
    'vii*': ['I', 'i'], # a pivot
}

just = ( 
    (1.0, 1.0),     # P1
    (16.0, 15.0),   # m2
    (9.0, 8.0),     # M2
    (6.0, 5.0),     # m3
    (5.0, 4.0),     # M3
    (4.0, 3.0),     # P4
    (45.0, 32.0),   # TT
    (3.0, 2.0),     # P5
    (8.0, 5.0),     # m6
    (5.0, 3.0),     # M6
    (9.0, 5.0),     # m7
    (15.0, 8.0),    # M7
)

terry = (
    (1.0, 1.0),     # P1  C
    (16.0, 15.0),   # m2  Db
    (10.0, 9.0),    # M2  D
    (6.0, 5.0),     # m3  Eb
    (5.0, 4.0),     # M3  E
    (4.0, 3.0),     # P4  F
    (64.0, 45.0),   # TT  Gb
    (3.0, 2.0),     # P5  G
    (8.0, 5.0),     # m6  Ab
    (27.0, 16.0),   # M6  A 
    (16.0, 9.0),    # m7  Bb
    (15.0, 8.0),    # M7  B
)

young = (
    (1.0, 1.0),       # P1 0
    (567.0, 512.0),   # m2 1
    (9.0, 8.0),       # M2 2
    (147.0, 128.0),   # m3 3
    (21.0, 16.0),     # M3 4
    (1323.0, 1024.0), # P4 5
    (189.0, 128.0),   # TT 6
    (3.0, 2.0),       # P5 7
    (49.0, 32.0),     # m6 8
    (7.0, 4.0),       # M6 9
    (441.0, 256.0),   # m7 10
    (63.0, 32.0),     # M7 11
)

louis = (
    (1.0, 1.0),       # P1 1.0
    (1.0, 1.0),       # m2 1.0
    (9.0, 8.0),       # M2 1.125
    (9.0, 8.0),       # m3 1.125
    (5.0, 4.0),       # M3 1.25
    (5.0, 4.0),       # P4 1.25
    (3.0, 2.0),       # TT 1.5
    (3.0, 2.0),       # P5 1.5
    (8.0, 5.0),       # m6 1.6
    (7.0, 4.0),       # M6 1.75
    (9.0, 5.0),       # m7 1.8
    (9.0, 5.0),       # M7 1.8
)


## scale mappings
# standard
major = (0, 2, 4, 5, 7, 9, 11)
minor = (0, 2, 3, 5, 7, 8, 10)

# notated as semitone deviations from 
# a major scale for readability
scales = {
    'major': major,
    'minor': minor,
    'super_locrian': (0, 2-1, 4-1, 5-1, 7-1, 9-1, 11-1),
    'neapolitan_minor': (0, 2-1, 4-1, 5, 7, 9-1, 11),
    'neapolitan_major': (0, 2-1, 4-1, 5, 7, 9, 11),
    'oriental': (0, 2-1, 4, 5, 7-1, 9, 11-1),
    'double_harmonic': (0, 2-1, 4, 5, 7, 9-1, 11),
    'enigmatic': (0, 2-1, 4, 5+1, 7+1, 9+1, 11),
}

# Maps to chromatic ratio lists above
notes = { 
        'a': 0,
        'a#': 1,
        'bb': 1, 
        'b': 2,
        'c': 3, 
        'c#': 4, 
        'db': 4, 
        'd': 5, 
        'd#': 6, 
        'eb': 6, 
        'e': 7, 
        'f': 8, 
        'f#': 9, 
        'gb': 9, 
        'g': 10, 
        'g#': 11, 
        'ab': 11, 
        }

midi_notes = {
    'a': 21,
    'a#': 22,
    'bb': 22,
    'b': 23, 
    'c': 24, 
    'c#': 25,
    'db': 25,
    'd': 26,
    'd#': 27,
    'eb': 27,
    'e': 28, 
    'f': 29, 
    'f#': 30, 
    'gb': 30, 
    'g': 31, 
    'g#': 32,
    'ab': 32,
}

def mtof(midi_note):
    return 2**((midi_note-69)/12.0) * 440

def ftom(freq):
    return math.log(freq / 440.0, 2) * 12 + 69

def ftomi(freq):
    return int(round(ftom(freq)))

def extractPitchClass(pitch):
    parsed = re.match('([a-zA-Z]#?b?)(\d+)', pitch)

    try:
        pitch_class = parsed.group(1)
    except AttributeError:
        return pitch

    return pitch_class

def ptom(pitch):
    pitch_class = extractPitchClass(pitch.lower())
    register = int(pitch[-1])
    midi_note = midi_map[pitch_class]
    midi_note += register * 12

    return midi_note

def edo(degree, divs=12):
    return 2**(degree/float(divs))

def edoRatios(divs=12):
    return [ (edo(i, divs), 1.0) for i in range(1, divs+1) ]

def edoScale(divs=12):
    return [ edo(i, divs) for i in range(1, divs+1) ]

def nti(note):
    """ Note to index
            returns the index of enharmonic note names
            or False if not found
    """
    return notes.get(note, False)

def ntf(note, octave=None, ratios=None):
    """ Note to freq 
    """
    if re.match('[a-zA-Z]#?b?\d+', note) is not None:
        parsed = re.match('([a-zA-Z]#?b?)(\d+)', note)
        note = parsed.group(1)
        octave = int(parsed.group(2))

    if ratios is None:
        ratios = terry

    if octave is None:
        octave = 4

    note_index = nti(note)
    mult = ratios[note_index][0] / ratios[note_index][1]

    if note_index >= 3:
        octave -= 1

    return mult * a0 * 2.0**octave

def stf(index):
    degree = index % 24
    octave = index / 24

    return (2 ** (degree / 24.0)) * (a0 / 4.0) * (2.0 ** octave)

def mtf(midi_note):
    return 2**((midi_note - 69) / 12.0) * 440.0

def fts(freq):
    # Try to find closest eq temp freq to input
    # Generate entire range of possible eq temp freqs
    all_freq = [ stf(index) for index in range(2**8) ]

    count = 0
    cfreq = 0
    while freq > cfreq and count < 2**8:
        cfreq = all_freq[count]
        count = count + 1

    count = count - 1

    return count % 55

def nts(note, octave):
    octave = octave if octave >= -2 else -2
    octave = octave if octave <= 8 else 8 

    degree = notes[note] * 2

    degree = degree + ((octave + 2) * 24) 

    return degree

def getscale(degrees, root=261.63, ratios=None, scale=None):
    if ratios is None:
        ratios = terry

    if scale is None:
        scale = major

    multipliers = []

    for degree in degrees:
        base = scale[(degree - 1) % len(scale)]
        octave = degree / (len(scale) + 1)

        multiplier = ratios[base][0] / ratios[base][1]
        multiplier = multiplier * 2**octave

        multipliers += [ multiplier ]

    return [ root * m for m in multipliers ]

def fromdegrees(scale_degrees=None, octave=2, root='c', scale=None, ratios=None):
    #major = (0, 2, 4, 5, 7, 9, 11)
    if scale_degrees is None:
        scale_degrees = [1,3,5]

    if ratios is None:
        ratios = terry

    if scale is None:
        scale = major

    freqs = []
    root = ntf(root, octave)

    for index, degree in enumerate(scale_degrees):
        # strings are okay
        degree = int(degree)

        # register offset 0+
        register = degree // (len(scale) + 1)

        chromatic_degree = scale[(degree - 1) % len(scale)]
        ratio = ratios[chromatic_degree]
        freqs += [ root * (ratio[0] / ratio[1]) * 2**register ]

    return freqs

def getQuality(name):
    quality = '-' if name.islower() else '^'
    quality = '*' if re.match(match_roman + '[*]', name) is not None else quality

    return quality

def getExtension(name):
    return re.sub(match_roman + '[/*+]?', '', name)

def getIntervals(name):
    quality = getQuality(name)
    extension = getExtension(name)

    chord = base_qualities[quality]

    if extension != '':
        chord = chord + extensions[extension]

    return chord

def getFreqFromChordName(name, root=440, octave=3, ratios=just):
    index = getChordRootIndex(name)
    freq = ratios[index]
    freq = root * (freq[0] / freq[1]) * 2**octave

    return freq

def stripChord(name):
    root = re.sub('[#b+/*^0-9]+', '', name)
    return root.lower()

def getChordRootIndex(name):
    root = chord_romans[stripChord(name)]

    if '#' in name:
        root += 1

    if 'b' in name:
        root -= 1

    return root % 12

def addIntervals(a, b):
    a = intervals[a]
    b = intervals[b]
    c = a + b
    for interval, index in intervals.items():
        if c == index:
            c = interval

    return c

def getRatioFromInterval(interval, ratios):
    try:
        index = intervals[interval]
    except IndexError:
        log('Interval out of range, doh. Here have a P1')
        index = 0

    try:
        ratio = ratios[index]
        ratio = ratio[0] / ratio[1]
    except IndexError:
        base_index = index % 12
        ratio = ratios[base_index]
        ratio = ratio[0] / ratio[1]

        register = (index - base_index) / 12
        ratio *= 2**register

    return ratio

def nextChord(name):
    name = stripChord(name)
    return random.choice(progressions[name])

def chord(name, key=None, octave=3, ratios=None):
    if key is None:
        key = default_key

    if ratios is None:
        ratios = terry

    key = ntf(key, octave, ratios)
    root = ratios[getChordRootIndex(name)]
    root = key * (root[0] / root[1])

    name = re.sub('[#b]+', '', name)
    chord = getIntervals(name)
    chord = [ getRatioFromInterval(interval, ratios) for interval in chord ]
    chord = [ root * ratio for ratio in chord ]

    return chord

def chords(names, key=None, octave=3, ratios=just):
    if key is None:
        key = default_key

    return [ chord(name, key, octave, ratios) for name in names ]

def fitScale(freq, scale):
    # Thanks to @kennytm from stack overflow for this <3 <3
    return min(scale, key=lambda x:abs(x-freq))

def fit(freq, low=20, high=20000, get_change=False):
    """ fit the given freq within the given freq range, 
    by transposing up or down octaves """

    # need to fit at least an octave
    if high < low * 2:
        high = low * 2

    def shift(freq, low, high, octave_shift=0):
        if freq < low:
            return shift(freq * 2, low, high, octave_shift + 1)

        if freq > high:
            return shift(freq * 0.5, low, high, octave_shift - 1)

        return freq, octave_shift

    freq, octave = shift(freq, low, high)

    if octave == 0:
        mult = 1
    elif octave > 0:
        mult = 2**octave
    elif octave < 0:
        mult = 1.0 / (2**abs(octave))

    if get_change == True:
        return freq, mult
    else:
        return freq



cdef double[:] _wavetable(int, int)
cdef double[:] _window(int, int)
cdef double[:] _adsr(int framelength, int attack, int decay, double sustain, int release)

cdef class Wavetable:
    cdef public double[:] data

cdef int SINE
cdef int COS
cdef int TRI
cdef int SAW
cdef int PHASOR
cdef int RSAW
cdef int HANN
cdef int HAMM
cdef int BLACK
cdef int BLACKMAN
cdef int BART
cdef int BARTLETT
cdef int KAISER
cdef int SQUARE
cdef int RND

cdef int LEN_WINDOWS
cdef int *ALL_WINDOWS
cdef int LEN_WAVETABLES
cdef int *ALL_WAVETABLES




import glob
import os

def parse_cents(tuning):
    try:
        return 2**(float(tuning) / 1200)
    except (TypeError, ValueError):
        return None

def parse_ratio(tuning):
    if '/' in tuning:
        try:
            ratio = [ t for t in tuning.split('/') if t != '' ][:2]
            return int(ratio[0])/int(ratio[1])
        except (TypeError, ValueError):
            return None

    try:
        return int(tuning)
    except (TypeError, ValueError):
        return None

def import_file(filename):
    if '.scl' not in filename:
        raise ValueError('This doesn\'t look like a Scala file')

    with open(filename, encoding='latin-1') as tuning_file:
        description = None
        scale_length = None
        tunings = []
        for line in tuning_file:
            if line[0] == '!':
                continue

            if description is None:
                description = line.strip()
                continue
            elif scale_length is None:
                try:
                    scale_length = int(line.strip())
                    continue
                except (TypeError, ValueError) as e:
                    raise ValueError('Invalid value for scale length') from e

            tuning = line.strip().split(' ')[0]

            if '.' in tuning:
                tuning = parse_cents(tuning)
            else:
                tuning = parse_ratio(tuning)

            if tuning is None:
                continue

            tunings += [ tuning ]

    return {
        'filename': filename, 
        'description': description, 
        'scale': [1] + tunings[:scale_length], 
    }


def import_directory(directory):
    scales = []
    for filename in glob.glob(os.path.join(directory, '*.scl')):
        try:
            scale = import_file(filename)
            scales += [ scale ]
        except Exception as e:
            print(e)
            continue

    return scales

if __name__ == '__main__':
    scales = import_directory('./scl/')
    print(len(scales))



""" Some shortcuts to doing boilerplate work with SoundBuffers
    which make live coding suck less, hopefully.
"""
import glob
import math
import multiprocessing as mp
import numpy as np
import random

cimport cython

from .soundbuffer import SoundBuffer
from .soundbuffer cimport SoundBuffer
from . cimport grains
cimport wavetables as wts

# Expose some C flags / constants to python
# FIXME might be faster to use newish cpdef enum defs? donno
SINE = wts.SINE
COS = wts.COS
TRI = wts.TRI
SAW = wts.SAW
PHASOR = wts.PHASOR
RSAW = wts.RSAW
HANN = wts.HANN
HAMM = wts.HAMM
BLACK = wts.BLACK
BLACKMAN = wts.BLACK
BART = wts.BART
BARTLETT = wts.BARTLETT
KAISER = wts.KAISER
SQUARE = wts.SQUARE
RND = wts.RND

# Just a shorthand for MS in scripts. 
# For example:
#
# >>> snd.cut(0.75, 32*dsp.MS)
#
# Which cuts, starting from 0.75 seconds in, 
# a 32 millisecond section from the SoundBuffer `snd`.
MS = 0.001

def mix(sounds):
    """ Mix a list of sounds into a new sound
    """
    out = SoundBuffer(length=1)
    for sound in sounds:
        out &= sound

    return out
    
cpdef wts.Wavetable wt(list values=None, int initwt=-1, int initwin=-1, int length=-1):
    return wts.Wavetable(values, initwt, initwin, length)

cpdef SoundBuffer stack(list sounds):
    cdef int channels = 0
    cdef int copied_channels = 0
    cdef int samplerate = sounds[0].samplerate
    cdef double length = 0
    cdef SoundBuffer sound

    for sound in sounds:
        channels += sound.channels
        if sound.dur > length:
            length = sound.dur

    cdef SoundBuffer out = SoundBuffer(length=length, channels=channels, samplerate=samplerate)

    cdef int channel_to_copy = 0
    cdef int current_sound_index = 0
    cdef int source_channel_to_copy = 0

    # TODO
    while copied_channels < channels:
        out.frames.base[:,channel_to_copy] = sounds[current_sound_index].frames.base[:,source_channel_to_copy]    
        copied_channels += 1

    return out

def join(sounds, overlap=None, channels=None, samplerate=None):
    """ Concatenate a list of sounds into a new sound
    """
    channels = channels or sounds[0].channels
    samplerate = samplerate or sounds[0].samplerate

    cdef double total_length = 0
    cdef SoundBuffer sound

    for sound in sounds:
        total_length += sound.dur
        if overlap is not None:
            total_length -= overlap

    out = SoundBuffer(length=total_length, channels=channels, samplerate=samplerate)

    pos = 0
    for sound in sounds:
        out.dub(sound, pos)
        pos += sound.dur
        if overlap is not None:
            pos -= overlap

    return out

def buffer(frames=None, length=-1, channels=2, samplerate=44100):
    """ Identical to `silence` -- creates an empty buffer of a given length
    """
    return SoundBuffer(frames=frames, length=length, channels=channels, samplerate=samplerate)

def read(frames, channels=2, samplerate=44100):
    """ Read a soundfile from disk and return a `SoundBuffer` with its contents
        Equiv to `snd = SoundBuffer(filename)`
    """
    if isinstance(frames, str):
        return SoundBuffer(filename=frames, channels=channels, samplerate=samplerate)

    return SoundBuffer(frames, channels=channels, samplerate=samplerate)

cpdef double rand(object lowvalue, object highvalue):
    if isinstance(lowvalue, tuple):
        lowvalue = rand(lowvalue[0], lowvalue[1])

    if isinstance(highvalue, tuple):
        highvalue = rand(highvalue[0], highvalue[1])

    return <double>random.triangular(lowvalue, highvalue)

def find(pattern, channels=2, samplerate=44100):
    """ Glob for files matching a given pattern and return a generator 
        that will `yield` each file as a `SoundBuffer`
    """
    for filename in glob.iglob(pattern, recursive=True):
        yield SoundBuffer(filename, channels=channels, samplerate=samplerate)

def pool(callback, reps=10, params=None, processes=4):
    out = []
    if params is None:
        params = [None]
    with mp.Pool(processes=processes) as process_pool:
        for result in [ process_pool.apply_async(callback, params[i % len(params)]) for i in range(reps) ]:
            out += [ result.get() ]

    return out

def fill(data, length):
    reps = math.ceil(length / len(data))
    if reps < 2:
        return data

    data = np.hstack([ data for _ in range(reps) ])
    return data[:length]




import numpy as np
cimport cython
from libc cimport math

ctypedef double (*get_sample_t)(double x, double y0, double y1, double y2, double y3)

@cython.boundscheck(False)
@cython.wraparound(False)
cdef double hermite_get_sample(double x, double y0, double y1, double y2, double y3):
    """ Taken from version #2 by James McCartney 
        http://musicdsp.org/showone.php?id=93
    """
    cdef double c0 = y1
    cdef double c1 = 0.5 * (y2 - y0)
    cdef double c3 = 1.5 * (y1 - y2) + 0.5 * (y3 - y0)
    cdef double c2 = y0 - y1 + c1 - c3

    return ((c3 * x + c2) * x + c1) * x + c0

cdef public double[:] _hermite(double[:] data, int length):
    cdef int fi = 0
    cdef double[:] out = np.zeros(length)
    cdef get_sample_t get_sample = hermite_get_sample
    cdef int datalen = len(data)
    cdef int i0, i1, i2 = 0
    cdef int pos = 0

    for fi in range(length):
        x = fi / float(length)
        pos = <int>(x * datalen)

        i0 = (pos - 1) % datalen
        i1 = (pos + 1) % datalen
        i2 = (pos + 2) % datalen

        out[fi] = get_sample(x, data[i0], data[pos], data[i1], data[i2])

    return out

cdef public double[:] _linear(double[:] data, int length):
    cdef double[:] out = np.zeros(length) 

    cdef int i = 0
    cdef int readindex = 0
    cdef int inputlength = len(data)
    cdef double phase = 0
    cdef double val, nextval, frac

    if inputlength <= 1:
        return out

    for i in range(length):
        readindex = <int>phase % inputlength

        val = data[readindex]

        try:
            nextval = data[readindex + 1]
        except IndexError:
            nextval = data[0]

        frac = phase - <int>phase

        val = (1.0 - frac) * val + (frac * nextval)

        out[i] = val

        phase += inputlength * (1.0 / length)

    return out

def linear(data, int length):
    if isinstance(data, list):
        data = np.asarray(data, dtype='d')
    return _linear(data, length)

def hermite(data, int length):
    if isinstance(data, list):
        data = np.asarray(data, dtype='d')
    return _hermite(data, length)




import numpy as np
import random
cimport cython
from .soundbuffer cimport SoundBuffer
from . cimport wavetables
from . import graph

cdef double MINDENSITY = 0.001

cpdef SoundBuffer go(SoundBuffer snd, 
                          double factor=30,
                          double density=1, 
                          double wet=1,
                          double minlength=0.01, 
                          double maxlength=0.06, 
                          double minclip=0.4, 
                          double maxclip=0.8, 
                          object win=None
                    ):
    if wet <= 0:
        return snd

    density = max(MINDENSITY, density)

    cdef double outlen = snd.dur + maxlength
    cdef SoundBuffer out = SoundBuffer(length=outlen, channels=snd.channels, samplerate=snd.samplerate)
    cdef wavetables.Wavetable window
    if win is None:
        window = wavetables.Wavetable(initwin=wavetables.HANN)
    else:
        window = wavetables.Wavetable(win)

    cdef double grainlength = random.triangular(minlength, maxlength)
    cdef double pos = 0
    cdef double clip
    cdef SoundBuffer grain

    while pos < outlen:
        grain = snd.cut(pos, grainlength)
        clip = random.triangular(minclip, maxclip)
        grain *= random.triangular(0, factor * wet)
        grain = grain.clip(-clip, clip)
        out.dub(grain * window.data, pos)

        pos += (grainlength/2) * (1/density)
        grainlength = random.triangular(minlength, maxlength)

    if wet > 0:
        out *= wet

    if wet < 1:
        out.dub(snd * abs(wet-1), 0)

    return out



#!/usr/bin/env python
from setuptools import setup

try:
    from Cython.Build import cythonize
    import numpy as np
    ext_modules = cythonize([
        'pippi/oscs.pyx', 
        'pippi/defaults.pyx', 
        'pippi/dsp.pyx', 
        'pippi/fx.pyx', 
        'pippi/soundbuffer.pyx', 
        'pippi/interpolation.pyx',
        'pippi/grains.pyx',
        'pippi/wavetables.pyx'
    ], include_path=[np.get_include()]) 

except ImportError:
    from setuptools.extension import Extension
    ext_modules = [
        Extension('pippi.oscs', ['pippi/oscs.c']), 
        Extension('pippi.dsp', ['pippi/dsp.c']), 
        Extension('pippi.fx', ['pippi/fx.c']), 
        Extension('pippi.grains', ['pippi/grains.c']), 
        Extension('pippi.defaults', ['pippi/defaults.c']), 
        Extension('pippi.soundbuffer', ['pippi/soundbuffer.c']), 
        Extension('pippi.interpolation', ['pippi/interpolation.c']),
        Extension('pippi.wavetables', ['pippi/wavetables.c']),
    ]

setup(
    name='pippi',
    version='2.0.0-beta-1',
    description='Computer music with Python',
    author='He Can Jog',
    author_email='erik@hecanjog.com',
    url='https://github.com/hecanjog/pippi',
    packages=['pippi', 'pippi.tune'],
    ext_modules=ext_modules, 
    install_requires=[
        'numpy', 
        'Pillow', 
        'PySoundFile'
    ], 
    keywords='algorithmic computer music', 
    python_requires='>=3',
    classifiers=[
        'Development Status :: 4 - Beta', 
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Other Audience', # Musicians and audio tinkerers, and of course anyone else...
        'License :: Public Domain', 
        'Natural Language :: English', 
        'Operating System :: POSIX', 
        'Operating System :: POSIX :: Linux', 
        'Operating System :: Microsoft :: Windows', 
        'Operating System :: MacOS', 
        'Programming Language :: Cython', 
        'Programming Language :: Python :: 3 :: Only', 
        'Programming Language :: Python :: 3.6', 
        'Topic :: Artistic Software', 
        'Topic :: Multimedia :: Sound/Audio', 
        'Topic :: Multimedia :: Sound/Audio :: Sound Synthesis', 
    ]
)



from .soundbuffer cimport SoundBuffer
from . cimport wavetables
from . cimport interpolation
from libc.stdlib cimport rand, RAND_MAX
import numpy as np
from cpython cimport array
import array

DEF MIN_DENSITY = 0.000001
DEF MIN_SPEED = 0.000001
DEF MIN_GRAIN_FRAMELENGTH = 8
DEF DEFAULT_WTSIZE = 4096
DEF DEFAULT_GRAINLENGTH = 60
DEF DEFAULT_MAXGRAINLENGTH = 80
DEF DEFAULT_SPEED = 1
DEF DEFAULT_MAXSPEED = 2
DEF DEFAULT_DENSITY = 1
DEF DEFAULT_MINDENSITY = 0.5
DEF DEFAULT_MAXDENSITY = 1.5

cdef class GrainCloud:
    def __init__(self, 
            SoundBuffer buf, 

            int win=-1,
            double[:] win_wt=None,
            int win_length=DEFAULT_WTSIZE,

            list mask=None,

            double freeze=-1, 
            int read_lfo=-1,
            object read_lfo_wt=None,
            double read_lfo_speed=1,
            int read_lfo_length=DEFAULT_WTSIZE,

            int speed_lfo=-1,
            double[:] speed_lfo_wt=None,
            int speed_lfo_length=DEFAULT_WTSIZE,
            double speed=DEFAULT_SPEED,
            double minspeed=DEFAULT_SPEED, 
            double maxspeed=DEFAULT_MAXSPEED, 

            int density_lfo=-1,
            double[:] density_lfo_wt=None,
            int density_lfo_length=DEFAULT_WTSIZE,
            double density_lfo_speed=1,
            double density=DEFAULT_DENSITY, 
            double mindensity=DEFAULT_MINDENSITY, 
            double maxdensity=DEFAULT_MAXDENSITY, 

            double grainlength=DEFAULT_GRAINLENGTH, 
            int grainlength_lfo=-1,
            double[:] grainlength_lfo_wt=None,
            int grainlength_lfo_length=DEFAULT_WTSIZE,
            double grainlength_lfo_speed=1,

            double minlength=DEFAULT_GRAINLENGTH,    # min grain length in ms
            double maxlength=DEFAULT_MAXGRAINLENGTH,    # max grain length in ms
            double spread=0,        # 0 = no panning, 1 = max random panning 
            double jitter=0,        # rhythm 0=regular, 1=totally random
            double amp=1, 
        ):

        if spread < 0:
            spread = 0
        elif spread > 1:
            spread = 1

        self.buf = buf
        self.channels = buf.channels
        self.samplerate = buf.samplerate
        self.spread = spread
        self.jitter = jitter
        self.freeze = freeze
        self.amp = amp

        cdef array.array cmask
        if mask is not None:
            cmask = array.array('i', mask)
            self.mask = cmask
    
        # Window is always a wavetable, defaults to Hann
        if win > -1:
            self.win = wavetables._window(win, win_length)
        elif win_wt is not None:
            self.win = win_wt
        else:
            self.win = wavetables._window(wavetables.HANN, win_length)
        self.win_length = len(self.win)

        # Read lfo is always a wavetable, defaults to phasor
        if self.freeze > -1:
            self.read_lfo = None
        elif read_lfo > -1:
            self.read_lfo = wavetables._window(read_lfo, read_lfo_length)
        elif read_lfo_wt is not None:
            #self.read_lfo = read_lfo_wt
            self.read_lfo = interpolation._linear(np.asarray(read_lfo_wt, dtype='d'), read_lfo_length)
        else:
            self.read_lfo = wavetables._window(wavetables.PHASOR, read_lfo_length)

        # 0 lfo speed freezes position
        if self.read_lfo is not None:
            if read_lfo_speed <= 0:
                read_lfo_speed = 0
            self.read_lfo_speed = read_lfo_speed
            self.read_lfo_length = len(self.read_lfo)

        # If speed_lfo < 0 and speed_lfo_wt is None, then use fixed speed
        # No transposition is done if speed == 1
        if speed_lfo > -1:
            self.speed_lfo = wavetables._window(speed_lfo, speed_lfo_length)
        elif speed_lfo_wt is not None:
            self.speed_lfo = speed_lfo_wt
        else:
            self.speed_lfo = None

        if self.speed_lfo is not None:
            self.speed_lfo_length = len(self.speed_lfo)

        if speed <= 0:
            speed = MIN_SPEED

        if minspeed <= 0:
            minspeed = MIN_SPEED

        if maxspeed <= 0:
            maxspeed = MIN_SPEED

        self.speed = speed
        self.minspeed = minspeed
        self.maxspeed = maxspeed

        self.density = density
        self.mindensity = mindensity
        self.maxdensity = maxdensity
        if density_lfo > -1:
            self.density_lfo = wavetables._window(density_lfo, density_lfo_length)
        elif density_lfo_wt is not None:
            self.density_lfo = density_lfo_wt
        else:
            self.density_lfo = None

        self.density_lfo_speed = density_lfo_speed
        if self.density_lfo is not None:
            self.density_lfo_length = len(self.density_lfo)

        if grainlength_lfo > -1:
            self.grainlength_lfo = wavetables._window(grainlength_lfo, grainlength_lfo_length)
            self.grainlength = -1
        elif grainlength_lfo_wt is not None:
            self.grainlength_lfo = grainlength_lfo_wt
        else:
            self.grainlength_lfo = None

        self.grainlength = grainlength
        self.minlength = minlength
        self.maxlength = maxlength
        self.grainlength_lfo_speed = grainlength_lfo_speed
        if self.grainlength_lfo is not None:
            self.grainlength_lfo_length = len(self.grainlength_lfo)
 
    def play(self, length=10):
        cdef int framelength = <int>(self.samplerate * length)
        cdef int input_length = len(self.buf)

        cdef int write_pos = 0              # frame position in output buffer for writing
        cdef int start = 0                  # grain start position in input buffer (frames)
        cdef int end = 0                    # grain end position in input buffer (frames)
        cdef double panpos = 0.5

        ###########
        # Read LFO
        ###########
        cdef double read_frac_pos = 0       # 0-1 pos in read buffer
        cdef double read_frac_pos_next = 0  # next value of 0-1 pos in read buffer for interpolation
        cdef double read_frac = 0           # fractional index for interpolation
        cdef double read_phase = 0
        cdef double read_phase_inc = (1.0 / framelength) * self.read_lfo_length
        cdef int read_pos = 0               # frame position in input buffer for reading
        cdef int read_index = 0

        ############
        # Speed LFO
        ############
        cdef double speed_frac_pos = 0       # 0-1 pos in speed buffer
        cdef double speed_frac_pos_next = 0  # next value of 0-1 pos in speed buffer for interpolation
        cdef double speed_frac = 0           # fractional index for interpolation
        cdef double speed_phase = 0
        cdef double speed_phase_inc = (1.0 / framelength) * self.speed_lfo_length
        cdef int speed_index = 0

        ##############
        # Density LFO
        ##############
        cdef double density_frac_pos = 0       # 0-1 pos in density buffer
        cdef double density_frac_pos_next = 0  # next value of 0-1 pos in density buffer for interpolation
        cdef double density_frac = 0           # fractional index for interpolation
        cdef double density_phase = 0
        cdef double density_phase_inc = (1.0 / framelength) * self.density_lfo_length
        cdef int density_pos = 0              # frame position in output buffer for writing
        cdef int density_index = 0

        ###################
        # Grain Length LFO
        ###################
        cdef double grainlength_frac_pos = 0       # 0-1 pos in grainlength buffer
        cdef double grainlength_frac_pos_next = 0  # next value of 0-1 pos in grainlength buffer for interpolation
        cdef double grainlength_frac = 0           # fractional index for interpolation
        cdef double grainlength_phase = 0
        cdef double grainlength_phase_inc = (1.0 / framelength) * self.grainlength_lfo_length
        cdef int grainlength_pos = 0              # frame position in output buffer for writing
        cdef int grainlength_index = 0
        cdef int grainlength = <int>(<double>self.samplerate * self.grainlength * 0.001)
        cdef int adjusted_grainlength = 0

        cdef int max_read_pos = input_length - grainlength

        cdef SoundBuffer out = SoundBuffer(length=length, channels=self.channels, samplerate=self.samplerate)

        cdef double maxspeed = <double>(input_length/4) / <double>min(self.grainlength, self.minlength)
        if self.maxspeed > maxspeed:
            # re-cap maxspeed based on grainlength
            self.maxspeed = maxspeed

        if self.speed > maxspeed:
            # re-cap maxspeed based on grainlength
            self.speed = maxspeed

        cdef int count = 0
        while write_pos < framelength - grainlength:
            # Grain length LFO
            if self.grainlength_lfo is not None:
                grainlength_phase = (<double>write_pos / framelength) * <double>self.grainlength_lfo_length * self.grainlength_lfo_speed
                grainlength_index = <int>grainlength_phase
                grainlength_frac = grainlength_phase - grainlength_index
                grainlength_frac_pos = self.grainlength_lfo[grainlength_index % self.grainlength_lfo_length]
                grainlength_frac_pos_next = self.grainlength_lfo[(grainlength_index+1) % self.grainlength_lfo_length]
                grainlength_frac_pos = (1.0 - grainlength_frac) * grainlength_frac_pos + grainlength_frac * grainlength_frac_pos_next
                grainlength_phase += grainlength_phase_inc

                if self.jitter > 0:
                    grainlength_frac_pos = grainlength_frac_pos * ((rand()/<double>RAND_MAX) * self.jitter + 1)

                grainlength = <int>(((grainlength_frac_pos * (self.maxlength - self.minlength)) + self.minlength) * (self.samplerate / 1000.0))
                max_read_pos = input_length - grainlength

            # Read LFO
            if self.read_lfo is None:
                if self.freeze < 0:
                    self.freeze = 0
                read_frac_pos = (self.samplerate * self.freeze) / <double>max_read_pos
                if read_frac_pos > 1:
                    read_frac_pos = 1
            else:
                read_phase = (<double>write_pos / (framelength - grainlength)) * <double>self.read_lfo_length * self.read_lfo_speed
                read_index = <int>read_phase
                read_frac = read_phase - read_index
                read_frac_pos = self.read_lfo[read_index % self.read_lfo_length]
                read_frac_pos_next = self.read_lfo[(read_index+1) % self.read_lfo_length]
                read_frac_pos = (1.0 - read_frac) * read_frac_pos + read_frac * read_frac_pos_next
                read_phase += read_phase_inc

            # Speed LFO
            if self.speed_lfo is not None:
                speed_phase = (<double>write_pos / (framelength - grainlength)) * <double>self.speed_lfo_length
                speed_index = <int>speed_phase
                speed_frac = speed_phase - speed_index
                speed_frac_pos = self.speed_lfo[speed_index % self.speed_lfo_length]
                speed_frac_pos_next = self.speed_lfo[(speed_index+1) % self.speed_lfo_length]
                speed_frac_pos = (1.0 - speed_frac) * speed_frac_pos + speed_frac * speed_frac_pos_next
                speed_phase += speed_phase_inc

            # Density LFO
            if self.density_lfo is not None:
                density_phase = (<double>write_pos / (framelength - grainlength)) * <double>self.density_lfo_length * self.density_lfo_speed
                density_index = <int>density_phase
                density_frac = density_phase - density_index
                density_frac_pos = self.density_lfo[density_index % self.density_lfo_length]
                density_frac_pos_next = self.density_lfo[(density_index+1) % self.density_lfo_length]
                density_frac_pos = (1.0 - density_frac) * density_frac_pos + density_frac * density_frac_pos_next
                density = (density_frac_pos * (self.maxdensity - self.mindensity)) + self.mindensity
                density_phase += density_phase_inc

            else:
                density = self.density

            if density < MIN_DENSITY:
                density = MIN_DENSITY

            if self.speed_lfo is not None:
                self.speed = speed_frac_pos * (self.maxspeed - self.minspeed) + self.minspeed

            #if self.mask is not None and self.mask[count] == 0:
            #    pass

            if self.speed != 1:
                adjusted_grainlength = <int>(grainlength * self.speed)
                if adjusted_grainlength > MIN_GRAIN_FRAMELENGTH:

                    start = <int>(read_frac_pos * (input_length-adjusted_grainlength))
                    grain = self.buf[start:start+adjusted_grainlength] 
                    grain = grain.speed(self.speed)
                else:
                    write_pos += MIN_GRAIN_FRAMELENGTH
                    continue
            else:
                start = <int>(read_frac_pos * max_read_pos)
                grain = self.buf[start:start+grainlength] 


            grain = grain * self.win

            if self.spread > 0:
                panpos = (rand()/<double>RAND_MAX) * self.spread + (0.5 - (self.spread * 0.5))
                grain = grain.pan(panpos)

            if write_pos + len(grain) < len(out):
                out._dub(grain * self.amp, write_pos)

            self.grains_per_sec = <double>self.samplerate / (<double>len(grain) / 2.0)
            self.grains_per_sec *= density

            write_pos += <int>(<double>self.samplerate / self.grains_per_sec)
            count += 1

        return out






""" Run this with `python -m pippi.benchmarks` to see some 
    timing information for synthesis etc.
"""
import timeit

init = """\
from pippi import oscs
osc = oscs.Osc()
"""

basic = """\
from pippi import oscs
osc = oscs.Osc()
out = osc.play({length})
"""

pwsine = """\
from pippi import oscs, wavetables, dsp
wavetable = wavetables.wavetable(dsp.SINE, {wtsize})
window = wavetables.window(dsp.SINE, {wtsize})
pulsewidth = 0.95
osc = oscs.Osc(wavetable, window=window, pulsewidth=pulsewidth)
out = osc.play({length})
"""

pwsinemod = """\
from pippi import oscs, wavetables, dsp
wavetable = wavetables.wavetable(dsp.SINE, {wtsize})
window = wavetables.window(dsp.SINE, {wtsize})
mod = wavetables.window(dsp.SINE, {wtsize})
pulsewidth = 0.95
osc = oscs.Osc(wavetable, window=window, mod=mod, pulsewidth=pulsewidth)
out = osc.play({length})
"""


if __name__ == '__main__':
    wtsize = 4096

    init_time = timeit.timeit(stmt=init, number=1000)
    print('init 1000x...')
    print(round(init_time, 2), '\n', init)

    run_times = [
        (1, 44100 * 10), 
        (10, 44100 * 10), 
        (100, 44100 * 10), 
    ]

    print('')
    print('basic')
    for nruns, length in run_times:
        print('%s %s second renders w/default settings...' % (nruns, round(length / 44100, 2)))
        basic_time = timeit.timeit(stmt=basic.format(length=length), number=nruns)
        t = round(basic_time, 2)
        print('total time %s\n%s %s\n%s seconds' % (t, round(t/nruns, 2), 'avg per run', basic.format(length=length)))
        print('')

    print('')
    print('pulsewidth w/sine window')
    for nruns, length in run_times:
        print('%s %s second renders pulsewidth w/sine window...' % (nruns, round(length / 44100, 2)))
        pwsine_time = timeit.timeit(stmt=pwsine.format(length=length, wtsize=wtsize), number=nruns)
        t = round(pwsine_time, 2)
        print('total time %s\n%s %s\n%s seconds' % (t, round(t/nruns, 2), 'avg per run', pwsine.format(length=length, wtsize=wtsize)))
        print('')

    print('')
    print('pulsewidth w/sine window & mod')
    for nruns, length in run_times:
        print('%s %s second renders pulsewidth w/sine window & mod...' % (nruns, round(length / 44100, 2)))
        pwsinemod_time = timeit.timeit(stmt=pwsinemod.format(length=length, wtsize=wtsize), number=nruns)
        t = round(pwsinemod_time, 2)
        print('total time %s\n%s %s\n%s seconds' % (t, round(t/nruns, 2), 'avg per run', pwsinemod.format(length=length, wtsize=wtsize)))
        print('')
    print('')

